//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FColorSelect.h - A simple button which allows to select the activated
//                  trace for a given color
//
// $Id: FColorSelect.h,v 1.9 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef FCOLORSELECT_H
#define FCOLORSELECT_H

#include <qpushbutton.h>
#include <qpopupmenu.h>
#include <qpixmap.h>

#include "FColors.h"

class FColorSelect : public QPushButton 
{
  Q_OBJECT

public:
  FColorSelect ( int num, QWidget * parent=0, const char * name=0 );

  ~FColorSelect();
  
  // get data about observed object
  bool       isActive()
  { return mainPopUp->isItemChecked( 0 ); }
  int        getPlayerNo()
  { return playerNo; }
  bool       isSideLeft()
  { return leftTeam; }

  // get flags of activated debugging output
  bool       hasDrawTrace()
  { return mainPopUp->isItemChecked( 3 ); }
  bool       hasWMTrace()
  { return mainPopUp->isItemChecked( 4 ); }
  bool       hasLogTrace()
  { return mainPopUp->isItemChecked( 5 ); }

  // some abbreviations for more convenient interfaces
  // check whether a color is a certain trace for something ...
  bool       isDrawTrace(int playerNo, bool leftTeam);
  bool       isWMTrace(int playerNo, bool leftTeam);
  bool       isLogTrace(int playerNo, bool leftTeam);

  const QColor *getColor() const
  { return &col; }
  
signals:
  void       stateChanged();

public slots:
  void       deactivate()
  { if(isActive()) toggleActive(); }
  void       activate()
  { if(!isActive()) toggleActive(); }
  void       setDrawTrace(const bool draw)
  { if (hasDrawTrace() != draw) toggleDrawTrace(); }
  void       setWMTrace(const bool WM)
  { if (hasWMTrace() != WM) toggleWM(); }
  void       setLogTrace(const bool log)
  { if (hasLogTrace() != log) toggleLog(); }

  void       setSidePlayer( const bool left, const int no);
  void       setTraces(const bool draw, const bool WM, const bool log)
  { setDrawTrace(draw); setWMTrace(WM); setLogTrace(log); }


protected slots:
  void       setPlayer0()
  { setSidePlayer( leftTeam, 0 ); recheckActive(); }
  void       setPlayer1()
  { setSidePlayer( leftTeam, 1 ); recheckActive(); }
  void       setPlayer2()
  { setSidePlayer( leftTeam, 2 ); recheckActive(); }
  void       setPlayer3()
  { setSidePlayer( leftTeam, 3 ); recheckActive(); }
  void       setPlayer4()
  { setSidePlayer( leftTeam, 4 ); recheckActive(); }
  void       setPlayer5()
  { setSidePlayer( leftTeam, 5 ); recheckActive(); }
  void       setPlayer6()
  { setSidePlayer( leftTeam, 6 ); recheckActive(); }
  void       setPlayer7()
  { setSidePlayer( leftTeam, 7 ); recheckActive(); }
  void       setPlayer8()
  { setSidePlayer( leftTeam, 8 ); recheckActive(); }
  void       setPlayer9()
  { setSidePlayer( leftTeam, 9 ); recheckActive(); }
  void       setPlayer10()
  { setSidePlayer( leftTeam, 10 ); recheckActive(); }
  void       setPlayer11()
  { setSidePlayer( leftTeam, 11 ); recheckActive(); }


  void       changeTeam()
  { setSidePlayer( !leftTeam, playerNo ); recheckActive(); }

  void       toggleActive();
  void       toggleDrawTrace();
  void       toggleWM();
  void       toggleLog();

  void       popup();

protected:
  int        playerNo;
  bool       leftTeam;

private:
  void       buildMenus();
  void       updateLabel();
  void       recheckActive() 
  { if (FColors::autoAct) activate(); }

  QPopupMenu *mainPopUp;
  QPopupMenu *playerNoMenu;
  QPixmap    *buttonPixm;
  QColor     col;
};

#endif
