//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FMagFactorDlg.h - small window for setting magnification factors
//                   in a fieldMap
//
// $Id: FMagFactorDlg.C,v 1.6 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FMagFactorDlg.moc"
#include "FMagFactorDlg.h"

#include "FieldMap.h"

FMagFactorDlg::FMagFactorDlg(QWidget *parent, const char * name)
  : QWidget( parent, name ), 
  playerEdit( new QLineEdit(this)),
  ballEdit( new QLineEdit(this)), 
  velEdit( new QLineEdit(this)),
  abortButton( new QPushButton("Abort", this )),
  setButton( new QPushButton("Set", this )),
  resetButton( new QPushButton("Reset", this )),
  fac1Button( new QPushButton("Factor 1", this )),
  dflPlayer( 1 ),
  dflBall( 1 ),
  dflSpeed( 1 )
{
  abortButton->resize(abortButton->sizeHint());
  setButton->resize(setButton->sizeHint());
  resetButton->resize(resetButton->sizeHint());
  fac1Button->resize(fac1Button->sizeHint());
    
  playerEdit->setText("1");
  playerEdit->resize(playerEdit->sizeHint());
  playerEdit->setMaxLength(5);
  playerEdit->setValidator(new QDoubleValidator(0.5, 20, 1, this));
  
  ballEdit->setText("1");
  ballEdit->resize(ballEdit->sizeHint());
  ballEdit->setMaxLength(5);
  ballEdit->setValidator(new QDoubleValidator(0.5, 20, 1, this));

  velEdit->setText("1");
  velEdit->resize(velEdit->sizeHint());
  velEdit->setMaxLength(5);
  velEdit->setValidator(new QDoubleValidator(0.5, 20, 1, this));

  int w(abortButton->width()+setButton->width()+
	resetButton->width() + fac1Button->width() + 4*5);
  int h(abortButton->height()+playerEdit->height()+
	ballEdit->height()+velEdit->height()+ 4*5);

  // set size and disallow resizing
  resize(w, h);
  setMaximumSize( size() );
  setMinimumSize( size() );

  QLabel *l= new QLabel("Player mag: ", this);
  l->move(0,0);
  l->resize(l->sizeHint());
  playerEdit->move(w-playerEdit->width(), 0);
  
  l= new QLabel("Ball Mag: ", this);
  l->move(0, playerEdit->height() + 5);
  l->resize(l->sizeHint());
  ballEdit->move(w-ballEdit->width(), l->y());
  
  l= new QLabel("Speed Mag: ", this);
  l->move(0, playerEdit->height() + 5 +velEdit->height() + 5);
  l->resize(l->sizeHint());
  velEdit->move(w-velEdit->width(), l->y());
    
  abortButton->move(0, h-abortButton->height());
  resetButton->move(abortButton->width() + 5, h-resetButton->height());
  fac1Button->move(abortButton->width() + 5 + resetButton->width() + 5,
		   h-setButton->height());
  setButton->move(abortButton->width() + 5 + resetButton->width() + 5
		  + fac1Button->width() + 5,  h-setButton->height());

  setMagSteps( dflPlayer, dflBall, dflSpeed );

  connect(abortButton, SIGNAL(clicked()), SLOT(hide()));
  connect(setButton, SIGNAL(clicked()), SLOT(setButtonPressed()));
  connect(resetButton, SIGNAL(clicked()), SLOT(resetButtonPressed()));
  connect(fac1Button, SIGNAL(clicked()), SLOT(fac1ButtonPressed()));
}

void FMagFactorDlg::setMagSteps( double player, double ball, double vel )
{
  char buf[20]="";
  sprintf(buf, "%5.1f", player);
  playerEdit->setText(buf);

  sprintf(buf, "%5.1f", ball);
  ballEdit->setText(buf);

  sprintf(buf, "%5.1f", vel);
  velEdit->setText(buf);
}

void FMagFactorDlg::setDflMagSteps( double player, double ball, double vel)
{
  dflPlayer= player;
  dflBall= ball;
  dflSpeed= vel;
}
  
void FMagFactorDlg::setButtonPressed()
{
  emit magSteps( atol(playerEdit->text()), 
		 atol(ballEdit->text()),
		 atol(velEdit->text()) );
  hide();
}

void FMagFactorDlg::fac1ButtonPressed()
{
  setMagSteps( 1, 1, 1 );
}

void FMagFactorDlg::resetButtonPressed()
{
  setMagSteps( dflPlayer, dflBall, dflSpeed );
}
