//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FMarkerWin.h - Window for maintaining marker list
//
// $Id: FMarkerWin.C,v 1.11 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FMarkerWin.moc"
#include "FMarkerWin.h"

#include <stdio.h>

FMarkerWin::FMarkerWin(const char *filename, QWidget * parent, 
		       const char * name )
  throw (Exception)
  : QWidget( parent, name ), listView( new QListView(this)),
    timeEdit(new QLineEdit(this)), 
    xEdit( new QLineEdit(this)), yEdit( new QLineEdit(this)),
    durtnEdit( new QLineEdit(this)),
    commEdit( new QLineEdit(this)),
    gotoButton( new QPushButton("Go to", this)),
    addButton( new QPushButton("Add", this)),
    resetButton( new QPushButton("Reset", this)),
    closeButton( new QPushButton("Close", this)),
    curMarker( fmlist.begin() ),
    time(0)
{
  timeEdit->setText("0");
  timeEdit->setMaxLength(6);
  timeEdit->resize(60, timeEdit->sizeHint().height());
  timeEdit->setValidator( new QIntValidator(-1, 10000, timeEdit));
  xEdit->setText("0");
  xEdit->setMaxLength(6);
  xEdit->resize(60, xEdit->sizeHint().height());
  xEdit->setValidator( new QDoubleValidator(-FIELD_LENGTH/2.0,
					   FIELD_LENGTH/2.0,2, xEdit ) );
  yEdit->setText("0");
  yEdit->setMaxLength(6);
  yEdit->resize(60, yEdit->sizeHint().height());
  yEdit->setValidator( new QDoubleValidator(-FIELD_WIDTH/2.0,
					   FIELD_WIDTH/2.0,2, yEdit ) );
  durtnEdit->setText("1");
  durtnEdit->setMaxLength(6);
  durtnEdit->resize(60, durtnEdit->sizeHint().height());
  durtnEdit->setValidator( new QIntValidator(-1, 10000, durtnEdit));
  
  commEdit->setText("");
  commEdit->setMaxLength(200);
  commEdit->resize(200, durtnEdit->sizeHint().height());

  int w(timeEdit->width()+xEdit->width()+yEdit->width()+
	durtnEdit->width()+commEdit->width()+4*5);

  gotoButton->resize(w/4, gotoButton->sizeHint().height());
  addButton->resize(w/4, addButton->sizeHint().height());
  resetButton->resize(w/4, resetButton->sizeHint().height());
  closeButton->resize(w/4, closeButton->sizeHint().height());

  connect(gotoButton, SIGNAL(clicked()), SLOT(gotoButtonClicked()));
  connect(addButton, SIGNAL(clicked()), SLOT(addButtonClicked()));
  connect(resetButton, SIGNAL(clicked()), SLOT(resetButtonClicked()));
  connect(closeButton, SIGNAL(clicked()), SLOT(closeButtonClicked()));
  
  connect(listView, SIGNAL(currentChanged(QListViewItem *)), 
	  SLOT(currentChanged(QListViewItem *)));
  connect(listView, SIGNAL(doubleClicked(QListViewItem *)), 
	  SLOT(doubleClicked(QListViewItem *)));
  connect(listView, SIGNAL(returnPressed(QListViewItem *)),
	  SLOT(doubleClicked(QListViewItem *)));

  listView->addColumn("time ", timeEdit->width());
  listView->addColumn("x    ", xEdit->width());
  listView->addColumn("y    ", yEdit->width());
  listView->addColumn("duration", durtnEdit->width());
  listView->addColumn("comment", commEdit->width());
  listView->setAllColumnsShowFocus(true);
  listView->resize(listView->sizeHint().width(), 200);
  
  resize(listView->height()+timeEdit->height()+addButton->height()+2*5,
	 w);
  setMinimumWidth(w);
  setMaximumWidth(w);
  
  setCaption("FUNSSELMon Markers");

  if (filename == 0)
    throw Exception("NULL string as filename argument for FMarkerWin");
  fname= new char[strlen(filename)+1];
  strcpy(fname, filename);

  // get saved marker list
  readFromFile();
}

FMarkerWin::~FMarkerWin() 
  throw (Exception)
{
  writeToFile();
  // collect all objects in marker list
  for (list<FieldMarker *>::iterator it= fmlist.begin(); 
       it!=fmlist.end(); it++)
    delete (*it);
}

const FieldMarker *FMarkerWin::getFirstMarker( int time )
{
  this->time= time;
  curMarker= fmlist.begin();

  if ((curMarker != fmlist.end()) && 
      ((*curMarker)->isInTime(time)))
    return *curMarker;
  else
    return getNextMarker();
}

const FieldMarker *FMarkerWin::getNextMarker()
{
  do {
  curMarker++;
  } while ((curMarker != fmlist.end()) && 
	   !((*curMarker)->isInTime(time)));

  return (curMarker == fmlist.end()) ? 0 : *curMarker;
}
  
void FMarkerWin::setValues( Vector2d pos, int time, int duration )
{
  char buf[10]= "";
  sprintf(buf, "%d", time);
  timeEdit->setText(buf);
  sprintf(buf, "%d", duration);
  durtnEdit->setText(buf);
  sprintf(buf, "%f", pos.getX());
  xEdit->setText(buf);
  sprintf(buf, "%f", pos.getY());
  yEdit->setText(buf);
}

void FMarkerWin::addMarker( Vector2d pos, int time, int duration,
			    const char *comm)
{
  FieldMarker *fm= new FieldMarker( pos, time, duration, comm );
  // check whether the marker is already in list:
  for (list<FieldMarker *>::const_iterator it= fmlist.begin();
       it != fmlist.end(); it++)
    if (*fm == **it) {
      delete fm;
      return;
    }
  
  char buftime[10], bufx[10], bufy[10], bufdurn[10];
  sprintf(buftime, "%5.5d", time);
  sprintf(bufx,    "%6.1f", pos.getX());
  sprintf(bufy,    "%6.1f", pos.getY());
  sprintf(bufdurn, "%5.5d", duration);
  new QListViewItem(listView, buftime, bufx, bufy, bufdurn, comm);
  fmlist.push_back( fm );
}

void FMarkerWin::clearList()
{
  list<FieldMarker *>::iterator it;
  for(it= fmlist.begin(); it != fmlist.end(); it++)
    delete (*it);
  fmlist.clear();

  listView->clear();
}
  
void FMarkerWin::currentChanged( QListViewItem *it)
{
  setValues( getPos(it), getTime(it), getDurtn(it) );
}

void FMarkerWin::doubleClicked( QListViewItem *it)
{
  emit selectedMarker( getTime(it) );
}

void FMarkerWin::gotoButtonClicked()
{
  if (QListViewItem *l= listView->currentItem())
    emit selectedMarker( getTime(l) );
}

void FMarkerWin::addButtonClicked()
{
  int t,d;
  double x,y;
  sscanf(timeEdit->text(), "%d", &t);
  sscanf(xEdit->text(), "%lf", &x);
  sscanf(yEdit->text(), "%lf", &y);
  sscanf(durtnEdit->text(), "%d", &d);

  addMarker( Vector2d(x, y), t, d, commEdit->text() );
}

void FMarkerWin::resetButtonClicked()
{
  timeEdit->setText("0");
  xEdit->setText("0");
  yEdit->setText("0");
  durtnEdit->setText("1");
  commEdit->setText("");
}

void FMarkerWin::closeButtonClicked()
{
  emit closedWindow();
  hide();
}

void FMarkerWin::resizeEvent( QResizeEvent * )
{
  int y( height()-addButton->height() );
  addButton->move(0, y);
  gotoButton->move(width()/2-gotoButton->width(), y);
  resetButton->move(width()/2, y);
  closeButton->move(width()-closeButton->width(), y);
  
  int y2(addButton->y()-5-timeEdit->height());
  timeEdit->move(0, y2);
  xEdit->move(timeEdit->x()+timeEdit->width()+5, y2);
  yEdit->move(xEdit->x()+xEdit->width()+5, y2);
  durtnEdit->move(yEdit->x()+yEdit->width()+5, y2);
  commEdit->move(durtnEdit->x()+durtnEdit->width()+5, y2);
  
  listView->move(0,0); 
  listView->resize(width(), y2-5);
}
  
int FMarkerWin::getTime(QListViewItem *it)
{
  int tmp(0);
  sscanf(it->text(0), "%d", &tmp);
  return tmp;
}

int FMarkerWin::getDurtn(QListViewItem *it)
{
  int tmp(0);
  sscanf(it->text(3), "%d", &tmp);
  return tmp;
}

Vector2d FMarkerWin::getPos(QListViewItem *it)
{
  double x,y;
  sscanf(it->text(1), "%lf", &x);
  sscanf(it->text(2), "%lf", &y);
  return Vector2d(x,y);
}

void FMarkerWin::readFromFile() 
  throw (Exception)
{
  ifstream ifs(fname);
  char buf[512], comm[201];
  int d, t;
  double x,y;
  
  while (ifs) {
    ifs.getline( buf, 512 );
    if (ifs) {
      if (sscanf(buf, "(fmark %d %d ( %lf , %lf ) \"%201[^\"]\" )", &t,
		 &d, &x, &y, comm) == 5) 
	addMarker(Vector2d(x,y), t, d, comm);
    }
  }
}

void FMarkerWin::writeToFile()
  throw (Exception)
{
  ofstream ofs(fname);
  
  if (ofs) {
    list<FieldMarker *>::iterator it;
    for (it= fmlist.begin(); it != fmlist.end(); it++)
      ofs << (**it) << endl;
  } else 
    throw Exception("Could not open marker file for writing "
		    + string(fname));
}
