//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FPlayerWin.h - Window for setting players to certain positions 
//
// $Id: FPlayerWin.C,v 1.4 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FPlayerWin.moc"
#include "FPlayerWin.h"

#include "constants.h"
#include "stdio.h"

FPlayerWin::FPlayerWin( QWidget * parent=0, const char * name=0 )
  : sideSel(new QComboBox(this)),plnum(new QComboBox(this)),
    xpos(new QLineEdit(this)), ypos( new QLineEdit(this)),
    ang(new QLineEdit(this)), 
    setButton(new QPushButton("Set", this)),
    resetButton(new QPushButton("Reset", this)),
    closeButton(new QPushButton("Close", this)),
    xlabel(new QLabel("pos: (", this)),
    ylabel(new QLabel(",", this)),
    anglabel(new QLabel(") ang:", this))
{
  sideSel->insertItem("Left",0);
  sideSel->insertItem("Right",1);
  sideSel->setCurrentItem(0);
  sideSel->resize(sideSel->sizeHint());

  plnum->insertItem("Player 1",0);
  plnum->insertItem("Player 2",1);
  plnum->insertItem("Player 3",2);
  plnum->insertItem("Player 4",3);
  plnum->insertItem("Player 5",4);
  plnum->insertItem("Player 6",5);
  plnum->insertItem("Player 7",6);
  plnum->insertItem("Player 8",7);
  plnum->insertItem("Player 9",8);
  plnum->insertItem("Player 10",9);
  plnum->insertItem("Player 11",10);
  plnum->setCurrentItem(0);
  plnum->resize(plnum->sizeHint());

  xpos->setText("  0.0");
  xpos->setMaxLength(6);
  xpos->resize(60, xpos->sizeHint().height());
  xpos->setValidator( new QDoubleValidator(-FIELD_LENGTH/2.0,
					   FIELD_LENGTH/2.0,2, xpos ) );

  ypos->setText("  0.0");
  ypos->setMaxLength(6);
  ypos->resize(60, ypos->sizeHint().height());
  ypos->setValidator( new QDoubleValidator(-FIELD_WIDTH/2.0,
					   FIELD_WIDTH/2.0,2, ypos ) );
  ang->setText("  0.0");
  xpos->setMaxLength(6);
  ang->resize(60, xpos->sizeHint().height());
  ang->setValidator( new QDoubleValidator(-360.0, 360.0,1, ang ) );

  xlabel->resize(xlabel->sizeHint());
  ylabel->resize(ylabel->sizeHint());
  anglabel->resize(anglabel->sizeHint());

  setButton->resize(setButton->sizeHint());
  connect( setButton, SIGNAL(clicked()), SLOT(setButtonPressed()));

  resetButton->resize(resetButton->sizeHint());
  connect( resetButton, SIGNAL(clicked()), SLOT(resetButtonPressed()));
  
  closeButton->resize(closeButton->sizeHint());
  connect( closeButton, SIGNAL(clicked()), SLOT(closeButtonPressed()));

  setFixedSize( xlabel->width()+xpos->width()+ylabel->width()
		+ ypos->width() + anglabel->width() + ang->width() + 5*5,
		sideSel->height()+ xpos->height()+ setButton->height()+ 3*5);
  
  setCaption("FUNSSELMon Set Player");
}

void FPlayerWin::setButtonPressed()
{
  double x,y,a;
  sscanf(xpos->text(), "%lf", &x);
  sscanf(ypos->text(), "%lf", &y);
  sscanf(ang->text(), "%lf", &a);
  emit setPlayerExec( sideSel->currentItem() == 0,
		      plnum->currentItem()+1,
		      Vector2d(x, y), DEG_TO_RAD(a));
}

void FPlayerWin::resetButtonPressed()
{
  xpos->setText("0");
  ypos->setText("0");
  ang->setText("0");
  plnum->setCurrentItem(0);
  sideSel->setCurrentItem(0);
}

void FPlayerWin::setPosn( Vector2d pos )
{
  char buf[10]="";
  sprintf(buf, "%5.2f", pos.getX());
  xpos->setText(buf);
  sprintf(buf, "%5.2f", pos.getY());
  ypos->setText(buf);
}

void FPlayerWin::resizeEvent ( QResizeEvent * )
{
  sideSel->move(0,0);
  plnum->move(sideSel->x()+sideSel->width()+5,0);
  
  int y(sideSel->height()+5);
  xlabel->move(0, y);
  xpos->move(xlabel->x()+xlabel->width()+5, y);
  ylabel->move(xpos->x()+xpos->width()+5, y);
  ypos->move(ylabel->x()+ylabel->width()+5, y);
  anglabel->move(ypos->x()+ypos->width()+5, y);
  ang->move(anglabel->x()+anglabel->width()+5, y);

  int y2(xpos->y()+ xpos->height() + 5);
  closeButton->move(0, y2); 
  resetButton->move((width()-resetButton->width())/2, y2);
  setButton->move(width()-setButton->width(), y2);  
}
