//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
// FieldMap.C - displays (parts of) the soccer field
// $Id: FieldMap.C,v 1.35 2001/09/11 10:08:37 goetz Exp $
// Time-stamp: <2001-08-31 14:01:14 goetz>
//////////////////////////////////////////////////////////////////////

#include "FieldMap.h"
#include "FieldMap.moc"
#include "FColors.h"

#include <qapplication.h>

#include <qsize.h>
#include <stdio.h>

#undef DRAWTRACE

//#define DRAWTRACE(output) cout <<"FieldMap: " << output << endl

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning: FieldMap debug trace DRAWTRACE is activated!
#endif

#define NUM_PLAYER_OBJS 100

const double FieldMap::DFLT_MAG( 5.0 );
const double FieldMap::DFLT_BORDER_SIZE( 3.0 );
const double FieldMap::DFLT_SPEED_MAG( 10.0 );
const double FieldMap::DFLT_PLAYER_MAG( 2.0 );
const double FieldMap::DFLT_BALL_MAG( 2.0 );

#ifdef QT2
#define QTString(str) QString(str)
#else
#define QTString(str) str
#endif

FieldMap::FieldMap( QWidget *parent, const char *name,
		    const double magnify,
		    const Vector2d &viewTopLeft,
		    const Vector2d &viewBotRight ) 
  : QWidget( parent, name ),
    magSetDlg( new FMagFactorDlg()),
    time(-1), pixmbuffer(0), dragging(false), 
    oldViewTopLeft(viewTopLeft), oldViewBotRight(viewBotRight),
    oldMag(magnify),
    mag_speed(DFLT_SPEED_MAG), mag_player(DFLT_PLAYER_MAG),
    mag_ball(DFLT_BALL_MAG),
    objs(0),
    leftTeamIcon(FColors::createTeamIcon(true)),
    rightTeamIcon(FColors::createTeamIcon(false))
{
  strcpy( status, "");

  setInitialDimensions();

  buildMenus();

  connect( magSetDlg, SIGNAL(magSteps( double, double, double)),
	   SLOT(magSteps( double, double, double)) );
}

FieldMap::~FieldMap()
{
  delete contextMenu;
  delete contextZoomMenu;
  delete contextAspRMenu;
  delete boxMenu;
  delete boxZoomMenu;
  delete pixmbuffer;
}

Vector2d FieldMap::getFC( const int x, const int y ) const
{
  // written in a more numerically stable fashion, not straightforward
  return Vector2d( (x * viewBotRight.getX() - x * viewTopLeft.getX())
		   / width() + viewTopLeft.getX(),
		   (y * viewBotRight.getY() - y * viewTopLeft.getY())
		   / height() + viewTopLeft.getY() );
}
  
QPoint FieldMap::getWP( const double x, const double y ) const
{
  // written in a more numerically stable fashion, not straightforward
  return QPoint( int( (x * width() - viewTopLeft.getX() * width() )
		      / (viewBotRight.getX() - viewTopLeft.getX()) ),
		 int( (y * height() - viewTopLeft.getY() * height() )
		      / (viewBotRight.getY() - viewTopLeft.getY()) ) );
}

bool FieldMap::showViewCone( bool leftTeam, int playerNo ) const 
{
  if (leftTeam)
    return showViewMenuLeft->isItemChecked(playerNo);
  else
    return showViewMenuRight->isItemChecked(playerNo);
}

void FieldMap::setView( const Vector2d & topLeft, const Vector2d & botRight )
{
  viewTopLeft= topLeft;
  viewBotRight= botRight;
  update();
}

void FieldMap::redrawObjs(const list<FieldObject *> *objlist)
{
  objs= objlist;
  buildPixmbuffer();
}

void FieldMap::magSteps( double player, double ball, double vel )
{
  mag_player= player;
  mag_ball= ball;
  mag_speed= vel;
}

void FieldMap::dflMagSteps( double player, double ball, double vel )
{
  mag_player= player;
  mag_ball= ball;
  mag_speed= vel;
  magSetDlg->setDflMagSteps( player, ball, vel );  
}


void FieldMap::correctAspR()
{
  Vector2d dim( viewBotRight - viewTopLeft );
  
  // find out which change is smaller and do this
  if ( abs( height() - int(width()* dim.getY()/dim.getX()) ) >
       abs( width() - int(height()* dim.getX()/dim.getY()) ))
    correctAspRHeight();
  else
    correctAspRWidth();

  emit changedSize();
}

void FieldMap::correctAspRWidth()
{
  Vector2d dim( viewBotRight - viewTopLeft );
  resize(width(), int(width()* dim.getY()/dim.getX()) );

  emit changedSize();
}

void FieldMap::correctAspRHeight()
{
  Vector2d dim( viewBotRight - viewTopLeft );
  resize(int(height()* dim.getX()/dim.getY()), height() );

  emit changedSize();
}

void FieldMap::enableAspRCorr( bool enabled )
{
  contextZoomMenu->setItemEnabled( 2, enabled );
}

void FieldMap::enableZoomThis( bool enabled )
{
  boxMenu->setItemEnabled( 0, enabled );
  contextZoomMenu->setItemEnabled( 1, enabled );
}

void FieldMap::setInitialDimensions()
{
  viewTopLeft= oldViewTopLeft;
  viewBotRight= oldViewBotRight;
  
  resize( QSize( (viewBotRight.getX() - viewTopLeft.getX()) * oldMag,
		 (viewBotRight.getY() - viewTopLeft.getY()) * oldMag ));
  buildPixmbuffer();
  emit changedSize();
  emit statusMsg(QTString("Window reset to initial dimensions."));
}

void FieldMap::zoomToWholeField()
{
  viewTopLeft= Vector2d(-FIELD_LENGTH/2 - DFLT_BORDER_SIZE, 
			-FIELD_WIDTH/2 - DFLT_BORDER_SIZE);
  viewBotRight= Vector2d(FIELD_LENGTH/2 + DFLT_BORDER_SIZE, 
			 FIELD_WIDTH/2 + DFLT_BORDER_SIZE);
  buildPixmbuffer();
  emit statusMsg(QTString("Zoomed to whole field."));
}

void FieldMap::boxZoomTo()
{
  QRect r=getBoxInOrder();
  setView( getFC( r.topLeft() ), getFC( r.bottomRight() ) );
  buildPixmbuffer();
  emit statusMsg(QTString("Zoomed to box."));
}

void FieldMap::boxZoomFact1()
{
  QRect r=getBoxInOrder();
  emit zoomWindow( getFC( r.topLeft() ), getFC( r.bottomRight() ), 
		   DFLT_MAG*1 );
}

void FieldMap::boxZoomFact2()
{
  QRect r=getBoxInOrder();
  emit zoomWindow( getFC( r.topLeft() ), getFC( r.bottomRight() ), 
		   DFLT_MAG*2 );
}

void FieldMap::boxZoomFact5()
{
  QRect r=getBoxInOrder();
  emit zoomWindow( getFC( r.topLeft() ), getFC( r.bottomRight() ), 
		   DFLT_MAG*5 );
}

void FieldMap::setMarker()
{
  emit markerSet ( getFC(mouseDownPos), mouseDownTime );
}

void FieldMap::dispFreekickLeft()
{
  emit sigFreekick( getFC(mouseDownPos), true );
}

void FieldMap::dispFreekickRight()
{
  emit sigFreekick( getFC(mouseDownPos), false );
}

void FieldMap::dispDropball()
{
  emit sigDropBall ( getFC(mouseDownPos) );
}

void FieldMap::setPlayerMenu()
{
  emit sigSetPlayer( getFC(mouseDownPos) );
}

void FieldMap::showMagStep()
{
  magSetDlg->setMagSteps( playermag(), ballmag(), speedmag() );
  magSetDlg->show();
}

void FieldMap::toggleKickRange()
{
  if (drawKickRange())
    showMenu->setItemChecked(0, false);
  else
    showMenu->setItemChecked(0, true);
}

void FieldMap::toggleViewConeLeft(int no)
{
  if (showViewMenu->isItemChecked(3)) {
    clearView();
    showViewMenuLeft->setItemChecked(no, true);
  } else
    showViewMenuLeft->setItemChecked(no, !showViewMenuLeft->isItemChecked(no));
}

void FieldMap::toggleViewConeRight(int no)
{
  if (showViewMenu->isItemChecked(3)) {
    clearView();
    showViewMenuRight->setItemChecked(no, true);
  } else
    showViewMenuRight->setItemChecked(no, !showViewMenuRight->isItemChecked(no));
}

void FieldMap::clearView()
{
  for (int i=0; i<11; i++) {
    showViewMenuLeft->setItemChecked(i, false);
    showViewMenuRight->setItemChecked(i, false);
  }
}

void FieldMap::toggleClearView()
{
  showViewMenu->setItemChecked(3,!(showViewMenu->isItemChecked(3)));
}

void FieldMap::mousePressEvent ( QMouseEvent * qme )
{
  DRAWTRACE("Coords: " << qme->pos().x() << " " << qme->pos().y() );
  
  mouseDownPos= qme->pos();
  mouseDragPos= qme->pos();
  mouseDownTime= time;
  if (qme->button() == LeftButton) 
    dragging= true;

  update();
}

void FieldMap::mouseMoveEvent ( QMouseEvent * qme )
{
  if (dragging) {
    mouseDragPos= qme->pos();

    if (mouseDragPos.x() < 0) 
      mouseDragPos.setX(0);
    else if (mouseDragPos.x() > width())
      mouseDragPos.setX(width());
    if (mouseDragPos.y() < 0)
      mouseDragPos.setY(0);
    else if (mouseDragPos.y() > height())
      mouseDragPos.setY(height());

    update();
    
    Vector2d vb(getFC(mouseDownPos)),
      ve(getFC(mouseDragPos)),
      d(ve-vb);
    
    sprintf(status, 
	    "Select : (%8.2f, %8.2f) -> (%8.2f, %8.2f) "
	    "dist: %8.2f, ang: %6.1f", 
	    vb.getX(), vb.getY(),
	    ve.getX(), ve.getY(),
	    d.getLength(), RAD_TO_DEG( d.getAngle() ) );
    emit statusMsg(QTString( status ));
  }
}

void FieldMap::mouseReleaseEvent ( QMouseEvent * qme )
{
  if (!dragging) {
    Vector2d v(getFC(mouseDownPos));
    if ( sprintf(status, 
		 "Position : (%8.2f, %8.2f)", 
		 v.getX(), v.getY() ) ) 
      emit statusMsg(QTString( status ));
    
    if (qme->button() == RightButton) 
      contextMenu->popup(qme->globalPos());
    else if (qme->button() == MidButton)
      foulMenu->popup(qme->globalPos());

  } else { // handle dragging events ...
    dragging= false;
    
    // only show boxMenu if a "box" is selected!
    if ( (mouseDragPos.x() != mouseDownPos.x())
	 && (mouseDragPos.y() != mouseDownPos.y())) 
      boxMenu->exec(qme->globalPos());
    emit statusMsg(QTString( "" ));
  }
  update();
}

void FieldMap::paintEvent ( QPaintEvent * pe )
{
  QPainter p;
  
  p.begin( this );
 
  if (dragging) {
    p.setPen( QPen( white, 2) );
    p.setBrush( NoBrush );
    p.setRasterOp( XorROP );
    p.drawRect( QRect( mouseDownPos, mouseDragPos ) );
  }
  p.end();
}

void FieldMap::resizeEvent( QResizeEvent * )
{
  buildPixmbuffer();
}

void FieldMap::buildMenus()
{
  foulMenu= new QPopupMenu;
  foulMenu->insertItem( "Drop Ball", this, 
			SLOT(dispDropball()) );
  foulMenu->insertItem( *leftTeamIcon, "Free kick left", this, 
			SLOT(dispFreekickLeft()) );
  foulMenu->insertItem( *rightTeamIcon, "Free kick right", this, 
			SLOT(dispFreekickRight()) );

  contextAspRMenu= new QPopupMenu;
  contextAspRMenu->insertItem( "intelligent", this,
			       SLOT(correctAspR()) );
  contextAspRMenu->insertItem( "keep width", this,
			       SLOT(correctAspRWidth()) );
  contextAspRMenu->insertItem( "keep height", this,
			       SLOT(correctAspRHeight()) );

  contextZoomMenu= new QPopupMenu;
  contextZoomMenu->insertItem( "original Viewport", this, 
			       SLOT(setInitialDimensions()), 0, 0 );
  contextZoomMenu->insertItem( "whole Field", this,
			       SLOT(zoomToWholeField()), 0, 1 );
  contextZoomMenu->insertItem( "Aspect ratio", 
			       contextAspRMenu, 2 );
  

  // create two Popups - one for each team
  showViewMenuLeft= new QPopupMenu;
  showViewMenuRight= new QPopupMenu;

  showViewMenuLeft->insertItem( "Player 1", this, SLOT(toggleViewConeL1()), 
				0, 1);
  showViewMenuLeft->insertItem( "Player 2", this, SLOT(toggleViewConeL2()), 
				0, 2);
  showViewMenuLeft->insertItem( "Player 3", this, SLOT(toggleViewConeL3()), 
				0, 3);
  showViewMenuLeft->insertItem( "Player 4", this, SLOT(toggleViewConeL4()), 
				0, 4);
  showViewMenuLeft->insertItem( "Player 5", this, SLOT(toggleViewConeL5()), 
				0, 5);
  showViewMenuLeft->insertItem( "Player 6", this, SLOT(toggleViewConeL6()), 
				0, 6);
  showViewMenuLeft->insertItem( "Player 7", this, SLOT(toggleViewConeL7()), 
				0, 7);
  showViewMenuLeft->insertItem( "Player 8", this, SLOT(toggleViewConeL8()), 
				0, 8);
  showViewMenuLeft->insertItem( "Player 9", this, SLOT(toggleViewConeL9()), 
				0, 9);
  showViewMenuLeft->insertItem( "Player 10", this, SLOT(toggleViewConeL10()), 
				0, 10);
  showViewMenuLeft->insertItem( "Player 11", this, SLOT(toggleViewConeL11()), 
				0, 11);

  showViewMenuRight->insertItem( "Player 1", this, SLOT(toggleViewConeR1()), 
				0, 1);
  showViewMenuRight->insertItem( "Player 2", this, SLOT(toggleViewConeR2()), 
				0, 2);
  showViewMenuRight->insertItem( "Player 3", this, SLOT(toggleViewConeR3()), 
				0, 3);
  showViewMenuRight->insertItem( "Player 4", this, SLOT(toggleViewConeR4()), 
				0, 4);
  showViewMenuRight->insertItem( "Player 5", this, SLOT(toggleViewConeR5()), 
				0, 5);
  showViewMenuRight->insertItem( "Player 6", this, SLOT(toggleViewConeR6()), 
				0, 6);
  showViewMenuRight->insertItem( "Player 7", this, SLOT(toggleViewConeR7()), 
				0, 7);
  showViewMenuRight->insertItem( "Player 8", this, SLOT(toggleViewConeR8()), 
				0, 8);
  showViewMenuRight->insertItem( "Player 9", this, SLOT(toggleViewConeR9()), 
				0, 9);
  showViewMenuRight->insertItem( "Player 10", this, SLOT(toggleViewConeR10()), 
				0, 10);
  showViewMenuRight->insertItem( "Player 11", this, SLOT(toggleViewConeR11()), 
				0, 11);
    
  // link them together with common stuff
  showViewMenu= new QPopupMenu;
  showViewMenu->insertItem("Left Team", showViewMenuLeft);
  showViewMenu->insertItem("Right Team", showViewMenuRight);
  showViewMenu->insertItem("no ViewCones", this, SLOT(clearView()), 0, 2);
  showViewMenu->insertItem("only one ViewCone", this, SLOT(toggleClearView()), 0, 3);
  showViewMenu->setItemChecked( 3, true );
  
  showMenu= new QPopupMenu;
  showMenu->insertItem("Show kickrange", this,
		       SLOT(toggleKickRange()), 0, 0);
  showMenu->insertItem("Show viewcone ...", showViewMenu);
  
  contextMenu= new QPopupMenu;
  contextMenu->insertItem("Zoom", contextZoomMenu);
  contextMenu->insertItem("Marker ...", this,
			  SLOT(setMarker()) );
  contextMenu->insertItem("SetPlayer ...", this, SLOT(setPlayerMenu()));
  contextMenu->insertItem("SetMagSteps ...", this, SLOT(showMagStep()));
  contextMenu->insertItem("Fouls", foulMenu );
  contextMenu->insertItem("Show ...", showMenu );

  boxZoomMenu= new QPopupMenu;
  boxZoomMenu->insertItem("Zoom 1:1", this,
			  SLOT( boxZoomFact1() ) );
  boxZoomMenu->insertItem("Zoom 2:1", this,
			  SLOT( boxZoomFact2() ) );
  boxZoomMenu->insertItem("Zoom 5:1", this,
			  SLOT( boxZoomFact5() ) );

  boxMenu= new QPopupMenu;
  boxMenu->insertItem( "zoom to box", this, SLOT( boxZoomTo()), 0, 0 );
  boxMenu->insertItem( "Open Zoomwindow", boxZoomMenu );
}

void FieldMap::buildPixmbuffer()
{
  DRAWTRACE("Building buffer ... ");

  if (objs == 0) {
    DRAWTRACE("No object list yet - nothing to do!");
    return;
  }

  delete pixmbuffer;
  pixmbuffer= new QPixmap( width(), height() );

  pixmbuffer->fill( FColors::background );

  QPainter p;
  
  p.begin( pixmbuffer );

  p.setBackgroundColor( FColors::background );

  // use a temporary background for drawing ...
  FieldBackground().draw(p, this);
  
  list<FieldObject *>::const_iterator it;
  for(it= objs->begin(); 
      it != objs->end(); 
      it++)
    (*it)->draw( p, this );
  
  p.end();
  
  setBackgroundPixmap( *pixmbuffer );
  repaint(true);
}

QRect FieldMap::getBoxInOrder()
{
  return QRect( mouseDownPos, mouseDragPos).normalize();
}

