//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "FunsselIO.h"
#include "Base.h"
#include <constants.h>
#include <stdio.h>

#include "FunsselStrings.h"

// cwz-build-command: make FUNSSELMon

FunsselIO::FunsselIO(int port, const char *host) :
  curtime(-1),
  smsg(SoccerMessage::emptyMessage),
  sock(port, host, MAX_SERVER_MSG_SIZE)
{
}

void FunsselIO::wind(int w)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (wind %d))", w);
  sock.send(buf);
}

void FunsselIO::windRel(int w)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (wind %c%d))", (w > 0) ? 'f' : 'b', abs(w));
  sock.send(buf);
}

void FunsselIO::maxRunTime(int t)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (max_run %d))", t);
  sock.send(buf);
}

void FunsselIO::startServer()
{
  sock.send(FILTER_STRING " (server start))");
}

void FunsselIO::quit()
{
  sock.send(FILTER_STRING " (quit))");
}

void FunsselIO::poll()
{
  sock.send(FILTER_STRING " (poll))");
}

void FunsselIO::init()
{
  sock.send(FILTER_STRING " (dispinit))");
}

void FunsselIO::sendMsg(const SoccerMessage &m)
{
  sock.send( string(m) );
}

FunsselIO::Result FunsselIO::getNextMsg()
{
  string str = sock.poll();
  if (str == string())
    return NoMsg;
  if (!strncmp(str.data(), FILTER_STRING, FILTER_STRING_LEN)) {
    // ok, a filtered msg. Depack
    char *parse;
    srvtime = strtol(str.data() + FILTER_STRING_LEN + 1, &parse, 10);
    parse++;
    if (!strncmp(parse, "(error", 6)) {
      cerr << "got error from FUNSSEL: (" <<  parse << endl;
      return NoMsg;
    }
    if (!strncmp(parse, "(server time changed)", 21))
      return NewSrvTime;
    if (!strncmp(parse, "(dispeol)", 9))
      return MsgEol;
    if (!strncmp(parse, "(insert failed)", 15))
      return MsgDropped;
    if (!strncmp(parse, "(max_run ok)", 12))
      return NoMsg;
    if (!strncmp(parse, "(welcome)", 9))
      return NoMsg;
    if (!strncmp(parse, "(server started)", 15))
      return ServerStarted;
    if (!strncmp(parse, "(wind", 5)) {
      int  cont = 6;
      bool stat = false;
      parse += 6;
      if (strncmp(parse, "failed", 6)) {
	if (!strncmp(parse, "ok", 2)) {
	  cont = 2;
	  stat = true;
	}
	else {
	  cerr << "cannot parse wind response " << parse + 6 << endl;
	  return NoMsg;
	}
      }
      curtime = atol(parse + cont);
      return (stat) ? WindOk : WindFailed;
    }
    if (!strncmp(parse, "(dispmsg", 8)) {
      int scanned, len;
      int i_sender, i_recver;
      parse += 9;
      if (sscanf(parse, "%d %d %d %d %d %n", &smsg.time,
		 &smsg.durtn, &i_sender, &i_recver,
		 &len, &scanned) != 5) {
	cerr << "cannot parse dispmsg " << parse << endl;
	return NoMsg;
      }
      smsg.sender = (SoccerActor)i_sender;
      smsg.recver = (SoccerActor)i_recver;
      smsg.message.assign(parse + scanned, len);
      return SoccerMsg;
    }
  }
  // ignore everything else
  // but not quietly
  cerr << "FUNSSEL sent unwanted " << str << endl;
  return NoMsg;
}


