//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-///////////////////////////////////////////////////////////
//
// MainWidget.C - main window of FUNSSEL program
//
// $Id: MainWidget.C,v 1.60 2001/09/11 10:08:38 goetz Exp $
//
////////////////////////////////////////////////////////////////////////


//#define USESOCKTRACE

#ifdef USESOCKTRACE
#define SOCKTRACE(output) cout <<"Socket(t=" << curTime() << "): " << \
 output << endl
#warning: MainWidget debug trace SOCKTRACE is activated!
#else
#define SOCKTRACE(output)
#endif


#include "MainWidget.moc"
#include "MainWidget.h"

// include Pixmaps
#include "rew0icon.xpm"
#include "rew1icon.xpm"
#include "stopicon.xpm"
#include "fwd1icon.xpm"
#include "fwd0icon.xpm"
#include "fwdcicon.xpm"
#include "fwdtoicon.xpm"
#include "jmpicon.xpm"
#include "fwddelay.xpm"
#include "fwddelayi.xpm"
#include "fwddelayd.xpm"

#ifdef QT2
#define QTString(str) QString(str)
#else
#define QTString(str) str
#endif

// non-existent time-step far in the future for finding the end of 
// logfiles, games ...
#define ETERNITY 25000

// value for "No playback bound" - has to be negative
#define PLAYBACKNOMAX -100

// for definition of SHOWINFO_SCALE
namespace SServerParams {
#include "param.h"
};

MainWidget::MainWidget(int port, const char *host,
		       const char *mfilename,
		       QWidget * parent, const char * name)
  : QWidget( parent, name ),
    logwin( new FLogWindow ),
    playerwin( new FPlayerWin ),
    objs( 0 ),
    leftTeamIcon( FColors::createTeamIcon(true)),
    rightTeamIcon( FColors::createTeamIcon(false)),
    fio( port, host ), haveOldLog(true), serverStarted( false ),
    gameRunning( false ),
    playBackDelay(30), playBackCnt(0), playBackMax(PLAYBACKNOMAX),
    listDirty( true ), windowIsOpen( true )
{
  buildMenu();

  setupObjects();
  matchFieldMapSize();

  connect(logwin, SIGNAL(closedWindow()), SLOT(closedLogWin()));
  connect(playerwin, SIGNAL(closedWindow()), SLOT(closedPlayerWin()));
  connect(playerwin, SIGNAL(setPlayerExec( bool, int, Vector2d, Angle)),
	  SLOT(setPlayerExec( bool, int, Vector2d, Angle)));

  const char *n;
  if (mfilename == 0)
    n= Config::getValue("FMonDflMarkerFile");
  else
    n= mfilename;
  
  markerwin= new FMarkerWin(n);

  connect(markerwin, SIGNAL(closedWindow()), SLOT(closedMarkerWin()));
  connect(markerwin, SIGNAL(selectedMarker(int)), SLOT(windMarker(int)));

  smp= new SoccerMsgParser( this, markerwin, logwin );

  // initialisation: connect with funssel
  fio.init();
  // check for logfile: send wind request
  fio.wind(-1);
  
  emit statusMsg(QTString("Initialized. Waiting for FUNSSEL ..."));
}

MainWidget::~MainWidget()
{
  deleteList( objs );
  delete markerwin;
  delete logwin;
  delete playerwin;
}

int MainWidget::getAColorDraw( int playerNo, bool leftTeam )
{
  int nCSel( FColors::numAgentColors() );

  for (int i=0; i<nCSel; i++) 
    if (cSel[i]->isDrawTrace(playerNo, leftTeam))
      return i;
  return -1;
}

int MainWidget::getAColorWM( int playerNo, bool leftTeam )
{
  int nCSel( FColors::numAgentColors() );

  for (int i=0; i<nCSel; i++) 
    if (cSel[i]->isWMTrace(playerNo, leftTeam))
      return i;
  return -1;
}

int MainWidget::getAColorLog( int playerNo, bool leftTeam )
{
  int nCSel( FColors::numAgentColors() );

  for (int i=0; i<nCSel; i++) 
    if (cSel[i]->isLogTrace(playerNo, leftTeam))
      return i;
  return -1;
}
  
void MainWidget::markerSet ( const Vector2d pos, int time )
{
  markerwin->setValues( pos, time );
  showMarkerWin( true );
}

void MainWidget::zoomWindow ( const Vector2d topLeft, 
			      const Vector2d botRight, double mag )
{
  FieldMap *zw= new FieldMap( 0, 0, mag, topLeft, botRight);
  zw->show();
  connectFieldMap( zw );
  setMagSteps( zw );
  zoomWindows.push_back( zw );
  zw->redrawObjs( objs );
}

void MainWidget::sigFreekick ( const Vector2d pos, bool sideLeft )
{
  char buf[40]="";
  int side( sideLeft ? 1 : -1 );
  sprintf(buf, "(dispfoul %d %d %d)", (int)(pos.getX()*SHOWINFO_SCALE),
	  (int)(pos.getY()*SHOWINFO_SCALE), side);
  fio.sendMsg(SoccerMessage(curTime(), string(buf), 
			    FunsselMonitor, SoccerServer));
  emit statusMsg(QTString("Freekick sent to server"));
}

void MainWidget::sigDropBall ( const Vector2d pos )
{
  char buf[40]="";
  sprintf(buf, "(dispfoul %d %d 0)", (int)(pos.getX()*SHOWINFO_SCALE),
	  (int)(pos.getY()*SHOWINFO_SCALE));
  fio.sendMsg(SoccerMessage(curTime(), string(buf), 
			    FunsselMonitor, SoccerServer));
  emit statusMsg(QTString("Drop ball sent to server"));
}

void MainWidget::sigSetPlayer ( const Vector2d pos )
{
  playerwin->setPosn( pos );
  showPlayerWin( true );
}

void MainWidget::setPlayerExec( bool sideLeft, int num, Vector2d pos, 
				Angle ang)
{
  char buf[40]="";
  int side( sideLeft ? 1 : -1 );
  sprintf(buf, "(dispplayer %d %d %d %d %d)", 
	  side, num,
	  (int)(pos.getX()*SHOWINFO_SCALE),
	  (int)(pos.getY()*SHOWINFO_SCALE), int(ang));
  fio.sendMsg(SoccerMessage(curTime(), string(buf), 
			    FunsselMonitor, SoccerServer));  
  emit statusMsg(QTString("Set player cmd sent to server"));
}

void MainWidget::windMarker( int time )
{
  fio.wind( time );
  emit statusMsg(QTString("Wind sent"));
}

void MainWidget::windRew0()
{
  // first go to timestep 1, then to -1 on second click
  if (curTime() <= 1) {
    fio.wind( -1 );
  } else {
    fio.wind( 1 );
  }
  if (!serverStarted) gameRunning= false;
  emit statusMsg(QTString("Wind sent. Game stopped."));
}

void MainWidget::windRew1()
{
  fio.windRel( -1 );
  if (!serverStarted) gameRunning= false;
  emit statusMsg(QTString("Wind sent. Game stopped"));
}

void MainWidget::windFwd1()
{
  fio.windRel( 1 );
  if (!serverStarted) gameRunning= false;
  emit statusMsg(QTString("Wind sent. Game stopped"));
}

void MainWidget::windFwdMax()
{
  fio.wind( ETERNITY );
  fio.windRel(0);
  if (!serverStarted) gameRunning= false;
  emit statusMsg(QTString("Wind sent. Game stopped"));
}

void MainWidget::playStop()
{
  if (gameRunning){
    fio.maxRunTime( curTime() );
    gameRunning= false;
    emit statusMsg(QTString("Game stopped"));
  } else {
    fio.windRel(0);
    emit statusMsg(QTString("Update done"));
  }
}

void MainWidget::playStart()
{
  if (serverStarted) {
    // normal game mode
    fio.maxRunTime( ETERNITY );
    fio.windRel( 1 );
    gameRunning= true;
    emit statusMsg(QTString("Game started"));
  } else {
    gameRunning= true;
    // reset upper bound for playback
    playBackMax= PLAYBACKNOMAX;
    emit statusMsg(QTString("Playback started"));
  }
}

void MainWidget::kickOff()
{
  if (!serverStarted) {
    startServer();
  } else {
    fio.sendMsg(SoccerMessage(-1, string("(dispstart)"), 
			      FunsselMonitor, SoccerServer));
    emit statusMsg(QTString("Kick off cmd sent to server"));
  }
}

void MainWidget::startServer()
{
  if (serverStarted) {
    if (!killServerQuery())
      return;
  } else 
    if (haveOldLog && !deleteFileQuery()) 
      return;
  
  serverStarted= true;
  btnKickOff->setText("Kick Off!");
  playMode->setText("Server started");
  markerwin->clearList();
  fio.startServer();
  playStart();

  // request next message
  fio.poll();
}

void MainWidget::setPlayDelay()
{
  char buf[40];
  playBackDelay= atoi( destTime->text() );
  if (playBackDelay < 0) {
    playBackDelay= 0;
    emit statusMsg(QTString("Cannot set negative delay! Playback speed delay "
		   "set to 0."));
  } else {
  sprintf(buf, "Playback speed delay set to %d", playBackDelay);
  emit statusMsg(QTString(buf));
  }
}

void MainWidget::incPlayDelay()
{
  char buf[40];
  playBackDelay+= 5;
  sprintf(buf, "Playback speed delay set to %d", playBackDelay);
  emit statusMsg(QTString(buf));
}

void MainWidget::decPlayDelay()
{
  char buf[40];
  playBackDelay-= 5;
  if (playBackDelay < 0) {
    playBackDelay= 0;
    emit statusMsg(QTString("Cannot set negative delay! Playback speed delay "
		   "set to 0."));
  } else {
    sprintf(buf, "Playback speed delay set to %d", playBackDelay);
    emit statusMsg(QTString(buf));
  }
}

void MainWidget::wind()
{
  fio.wind( atoi( destTime->text() ) );
  gameRunning= false;
  emit statusMsg(QTString("Wind sent. Game stopped"));
}

void MainWidget::playMaxTime()
{
  if (serverStarted) {
    // behaviour during game
    fio.maxRunTime( atoi( destTime->text() ) );
    fio.windRel( ETERNITY );
    gameRunning= true;
    emit statusMsg(QTString("Game started"));
  } else {
    // behaviour during playback
    playBackMax= atoi( destTime->text() );
    gameRunning= true;
    fio.windRel(1);
    emit statusMsg(QTString("Game started with stop time."));
  }
}

void MainWidget::doRedraw()
{
  SOCKTRACE("doRedraw");
  if (!listDirty)
    return;
  // now that we have a new list, force a redraw of all FieldMaps
  mainMap->redrawObjs(objs);
  for (list<FieldMap *>::iterator it(zoomWindows.begin());
       it != zoomWindows.end(); it++)
    (*it)->redrawObjs(objs);
  listDirty= false;
}

void MainWidget::pollMessages()
{
  // try to get new msgs while server is stopped ...
  if (!serverStarted && gameRunning) {
    if (playBackCnt >= playBackDelay ) {
      SOCKTRACE("Going forward one step");
      
      // in playback mode check for playMaxTime
      if (playBackMax > PLAYBACKNOMAX) {
	// playMaxTime feature enabled ...
	if (curTime() < playBackMax) {
	  fio.windRel(1);
	} else {
	  playBackMax= PLAYBACKNOMAX;
	  gameRunning= false;
	}
      } else {
	// no playMaxTime
	fio.windRel(1);
      }
      playBackCnt= 0;
    } else
      playBackCnt++;
  }
  
  int res = fio.getNextMsg();
  // check for data from new time step
  // (after wind)
  if (fio.currentTime() != curTime()) {
    SOCKTRACE("Got new current time");
    doRedraw();
    setCurrentTime(fio.currentTime());
  }

  switch (res) {
  case FunsselIO::NewSrvTime:
    SOCKTRACE("Got new server time");
    setServerTime(fio.serverTime());
    fio.wind(fio.serverTime());
    doRedraw();
    break;
  case FunsselIO::ServerStarted:
    SOCKTRACE("Got \"Server Started\"");
    fio.sendMsg(SoccerMessage(-1, string("(dispinit version 2)"),
			      FunsselMonitor, SoccerServer));
    break;
  case FunsselIO::SoccerMsg: {
    SOCKTRACE("SoccerMsg.");

    FieldObject *f= smp->parseMessage(fio.currentMsg());
    if (f != 0 && f->type() == FieldObject::FOServer) {
      deleteList( objs );

      // secondly, clear log window if desired
      if (logwin->clearStep())
	logwin->clearList();
    }

    // if the list was deleted or not yet created, add basic stuff...
    if (objs == 0) 
      objs= basicList();
    
    if (f != 0) {
      // adding object to list ...
      objs->push_back( f );
      listDirty= true;
    }

    fio.poll();

  }  break;
  case FunsselIO::MsgEol:

    SOCKTRACE("MsgEol.");
    // graphical output
    doRedraw();

    break;
  case FunsselIO::WindOk: {

    SOCKTRACE("WindOk: " << fio.currentTime());
    
    setCurrentTime( fio.currentTime() );

    fio.poll();

    if ((curTime() == -1) && !serverStarted)
      emit statusMsg(QTString(""));
    
    break;
    }
  case FunsselIO::WindFailed:

    SOCKTRACE("WindFailed: " << fio.currentTime() );

    // special case: we have no logfile if time==-2
    if (fio.currentTime() == -2)
      haveOldLog= false;

    fio.poll();
    
    break;
  case FunsselIO::NoMsg:

    SOCKTRACE("NoMsg.");
    
    // nothing happened, so sleep for a while so we don't get too many
    // system resources
    usleep(15000);
    break;
  default:
    cerr << "Illegal result from FunsselIO::getNextMsg()" << endl;
  }
}

void MainWidget::showMarkerWin( bool show )
{
  windowMenu->setItemChecked(0, show);

  if (show)
    markerwin->show();
  else
    markerwin->hide();
}

void MainWidget::showPlayerWin( bool show )
{
  windowMenu->setItemChecked(1, show);

  if (show)
    playerwin->show();
  else
    playerwin->hide();
}

void MainWidget::showLogWin( bool show )
{
  windowMenu->setItemChecked(2, show);

  if (show)
    logwin->show();
  else
    logwin->hide();
}

void MainWidget::matchFieldMapSize()
{
  int w( mainMap->width()+20),
    h( mainMenu->height() + 5 +
       destTime->height() + 5 +
       leftTeam->height() + 5 + 
       mainMap->height() + 5 +
       cSel[0]->height() + 5 +
       statusBar->height() );
  resize( w, h );
}

inline int rightSide( QWidget *w)
{
  return w->x()+w->width();
}

void MainWidget::quitApplication()
{
  if (closeQuery()) {
    hide();
    fio.quit();
    windowIsOpen= false;
  }
}

void MainWidget::aboutBox()
{
  QMessageBox::about(this,"FUNSSEL (c) by MRB",
		     "FUNSSEL - Flexible Utility for \n"
		     "Network based Soccer Simulation \n"
		     "using Extended Language \n\n"
		     "Version: " VERSION "\n"
		     "(c) 2001 by Mainz Rolling Brains\n"
		     "RoboCup Team.\n\n"
		     "published under the terms of GPL 2\n\n"
		     "FUNSSEL written by Axel Arnold\n"
		     "FUNSSELMon written by Goetz Schwandtner\n"
		     "\nContact: funssel@rollingbrains.de\n"
		     "\nwww.rollingbrains.de");
}

void MainWidget::aboutQt()
{
  QMessageBox::aboutQt(this);
}

void MainWidget::setCurrentTime(int t)
{
    char buf[20];
    sprintf(buf, "%d", t );
    timeDisp->setText(buf);
    
    mainMap->setTime(t);
    for (list<FieldMap *>::iterator it= zoomWindows.begin();
	 it != zoomWindows.end(); it++)
      (*it)->setTime(t);
}

void MainWidget::setServerTime(int t)
{
    char buf[20];
    sprintf(buf, "S: %d", t );
    serverTime->setText(buf);
}


void MainWidget::resizeEvent ( QResizeEvent * )
{
  // mainMenu is automatically set to the top of the widget & resized

  // set widgets below menu bar:
  int fstLine(mainMenu->height()+5);
  btnKickOff->move(10, fstLine);
  rewBeg->move(rightSide(btnKickOff)+10, fstLine);
  rewStep->move(rightSide(rewBeg), fstLine);
  btnStop->move(rightSide(rewStep), fstLine);
  fwdStep->move(rightSide(btnStop), fstLine);
  fwdPlay->move(rightSide(fwdStep), fstLine);
  fwdMax->move(rightSide(fwdPlay), fstLine);
  destTime->move(rightSide(fwdMax)+10, fstLine);
  playTo->move(rightSide(destTime)+10, fstLine);
  jumpTo->move(rightSide(playTo), fstLine);
  btnfwddelay->move(rightSide(jumpTo)+5, fstLine);
  btnfwddelayi->move(rightSide(btnfwddelay), fstLine);
  btnfwddelayd->move(rightSide(btnfwddelayi), fstLine);

  // probabely edit is the biggest widget ?!
  int scdLine(destTime->y()+destTime->height()+5);
  // align the time display into the middle and the other
  // ones into the free space at both sides
  int wid( (width()-4*10-3*5-timeDisp->width()-playMode->width()
	    - serverTime->width())/2);
  leftTeam->setGeometry( 10, scdLine, wid, leftTeam->height());
  timeDisp->move( rightSide(leftTeam)+5, scdLine);
  playMode->move( rightSide(timeDisp)+10, scdLine);
  serverTime->move( rightSide(playMode)+10, scdLine);
  rightTeam->setGeometry( rightSide(serverTime)+5, scdLine,
			  wid, rightTeam->height() );

  // set lower bound widgets now: first status bar:
  statusBar->setGeometry(0,height()-statusBar->height(),
			 width(),statusBar->height());
  // then color traces buttons above:
  // make them all the same width
  int num( FColors::numAgentColors() ),
    btnWid((width()-10)/num), btnHgt(cSel[0]->height()),
    btnTop( statusBar->y()-btnHgt );
  for (int i=0; i<num; i++)
    cSel[i]->setGeometry(5+i*btnWid, statusBar->y()-btnHgt,
			 btnWid-5, btnHgt);
  
  // AAAIIIEEEEE - A MONSTER: our big FieldMap widget
  int mainY( leftTeam->y() + leftTeam->height() + 5),
    mainHgt( btnTop - mainY );
  mainMap->setGeometry( 5, mainY, width()-10, mainHgt );
}

void MainWidget::closeEvent ( QCloseEvent *e)
{
  if (closeQuery()) {
    e->accept();
    windowIsOpen= false;
  }
}

bool MainWidget::closeQuery()
{
  if (QMessageBox::information ( this, "FUNSSEL Question",
				 "Finish working with this "
				 "wonderful program?", "Yes", "No", 
				 0, 0, 1 ) == 0) {
    fio.quit();
    return true;
  } else
    return false;
}

bool MainWidget::deleteFileQuery()
{
 if (QMessageBox::information ( this, "FUNSSEL Question",
				"This will delete the old\n"
				"game logs and start a new game.",
				"I know", "Stop !",
				 0, 0, 1 ) == 0) {
    return true;
  } else
    return false;
}

bool MainWidget::killServerQuery()
{
 if (QMessageBox::information ( this, "FUNSSEL Question",
				"This will kill the running\n"
				"game and start a new server\n"
				"with a new logfile\n"
				"All information about the current\n"
				"game will be lost.",
				"I know", "Stop !",
				 0, 0, 1 ) == 0) {
    return true;
  } else
    return false;
}

void MainWidget::buildMenu()
{
  fileMenu= new QPopupMenu;
  fileMenu->insertItem("&Quit", this, SLOT( quitApplication() ), Key_Q );
  
  controlMenu= new QPopupMenu;
  controlMenu->insertItem(QPixmap(fwdcicon),"Sta&rt", this, 
			  SLOT(playStart() ), Key_P);
  controlMenu->insertItem(QPixmap(stopicon),"&Stop and update Screen", this, 
			  SLOT(playStop()), Key_S);
  controlMenu->insertSeparator();
  controlMenu->insertItem(QPixmap(rew0icon),"Wind to &begin", this, 
			  SLOT(windRew0()));
  controlMenu->insertItem(QPixmap(rew1icon),"Wind &back 1 step", this, 
			  SLOT(windRew1()), Key_B);
  controlMenu->insertItem(QPixmap(fwd1icon),"Wind &fwd 1 step", this, 
			  SLOT(windFwd1()), Key_F);
  controlMenu->insertItem(QPixmap(fwd0icon),"Wind to &end", this,
			  SLOT(windFwdMax()));
  controlMenu->insertSeparator();
  controlMenu->insertItem(QPixmap(fwdtoicon),"&Play until ...", this, 
			  SLOT(playMaxTime()), Key_U);
  controlMenu->insertItem(QPixmap(jmpicon), "&Jump to", this, 
			  SLOT(wind()), Key_Return);
  controlMenu->insertSeparator();
  controlMenu->insertItem(QPixmap(fwddelay), "Set &delay", this, 
			  SLOT(setPlayDelay()));
  controlMenu->insertItem(QPixmap(fwddelayi), "&Increase delay", this, 
			  SLOT(incPlayDelay()));
  controlMenu->insertItem(QPixmap(fwddelayd), "D&ecrease delay", this, 
			  SLOT(decPlayDelay()));
  controlMenu->insertSeparator();
  controlMenu->insertItem("&Kick off", this, 
			  SLOT(kickOff()), Key_K);
  controlMenu->insertItem("Start &new server", this, 
			  SLOT(startServer()));

  dispObjMenu= new QPopupMenu;
  dispObjMenu->insertItem(*leftTeamIcon, "&Left Team", this, 
			  SLOT(toggleShowLeft()), 0, 0);
  dispObjMenu->insertItem(*rightTeamIcon, "&Right Team", this, 
			  SLOT(toggleShowRight()), 0, 1);
  dispObjMenu->insertItem("&Player output", this, 
			  SLOT(toggleShowTrace()), 0, 2);
  dispObjMenu->insertSeparator();
  dispObjMenu->insertItem("Log &audio msgs", this, SLOT(toggleAudioMsg()), 
			  0, 3);
  dispObjMenu->insertItem("Log player &commands", this, 
			  SLOT(togglePlayerCmds()), 0, 4);
  dispObjMenu->insertSeparator();
  dispObjMenu->insertItem("Player &may control", this,
			  SLOT(togglePlayerMayControl()), 0, 5);
  dispObjMenu->insertSeparator();
  dispObjMenu->insertItem("&Display stamina values", this,
			  SLOT(toggleDisplayStamina()), 0, 6);
  dispObjMenu->insertItem("Display player &types", this,
			  SLOT(toggleDisplayPlayerType()), 0, 7);

  // switch on all objects in this menu
  for (int i=0; i<3; i++)
    dispObjMenu->setItemChecked(i, true);

  windowMenu= new QPopupMenu;
  windowMenu->insertItem("&Marker Window", this, SLOT(toggleMarkerWin()),
			 Key_M, 0);
  windowMenu->insertItem("&Set Player Window", this, SLOT(togglePlayerWin()),
			 0, 1);
  windowMenu->insertItem("&Logfile Window", this, SLOT(toggleLogWin()), 
			 Key_L, 2);

  aboutMenu= new QPopupMenu;
  aboutMenu->insertItem("&Copyright", this, SLOT(aboutBox()), Key_F1);
  aboutMenu->insertItem("about &Qt", this, SLOT(aboutQt()));

  mainMenu= new QMenuBar( this );
  mainMenu->insertItem("&File", fileMenu);
  mainMenu->insertItem("Game &control", controlMenu);
  mainMenu->insertItem("&Graphics", dispObjMenu);
  mainMenu->insertItem("&Window", windowMenu);
  mainMenu->insertItem("&About", aboutMenu);
  mainMenu->setSeparator( QMenuBar::InWindowsStyle );
}

void MainWidget::setupObjects()
{
  statusBar= new QStatusBar( this );

#ifdef QT2  
  QObject::connect( this, SIGNAL(statusMsg(const QString &)),
		    statusBar, SLOT(message(const QString &)));
#else
  QObject::connect( this, SIGNAL(statusMsg(const char *)),
		    statusBar, SLOT(message(const char *)));
#endif

  mainMap= new FieldMap(this);
  // keep view of whole field!!!
  mainMap->enableZoomThis( false );

  connect( mainMap, SIGNAL( changedSize()), 
	   SLOT(matchFieldMapSize()));
  connectFieldMap( mainMap );
  setMagSteps( mainMap, true );
  
  leftTeam= new QLabel("0 left Team", this);
  leftTeam->setBackgroundColor( FColors::leftTeam.color() );
  leftTeam->setAlignment(AlignLeft | AlignVCenter);
  leftTeam->resize(leftTeam->sizeHint());
  rightTeam= new QLabel("0 right Team", this);
  rightTeam->setBackgroundColor( FColors::rightTeam.color() );
  rightTeam->setAlignment(AlignRight | AlignVCenter);
  rightTeam->resize(rightTeam->sizeHint());  

  timeDisp= new QLabel(" -2    ", this);
  timeDisp->setAlignment(AlignCenter);
  timeDisp->resize(timeDisp->sizeHint());

  playMode= new QLabel("   Start Server !   ", this);
  playMode->setAlignment(AlignCenter);
  playMode->resize(playMode->sizeHint());

  serverTime= new QLabel("S: -2    ", this);
  serverTime->setAlignment(AlignCenter);
  serverTime->resize(serverTime->sizeHint());

  btnKickOff= new QPushButton("Start Server!", this);
  btnKickOff->resize(btnKickOff->sizeHint());
  connect(btnKickOff, SIGNAL(clicked()), SLOT(kickOff()) );

  rewBeg= new QPushButton(this);
  rewBeg->setPixmap(QPixmap(rew0icon));
  rewBeg->resize(rewBeg->sizeHint());
  connect(rewBeg, SIGNAL(clicked()), SLOT(windRew0()) );
  rewStep= new QPushButton(this);
  rewStep->setPixmap(QPixmap(rew1icon));
  rewStep->resize(rewStep->sizeHint());
  connect(rewStep, SIGNAL(clicked()), SLOT(windRew1()) );
  btnStop= new QPushButton(this);
  btnStop->setPixmap(QPixmap(stopicon));
  btnStop->resize(btnStop->sizeHint());
  connect(btnStop, SIGNAL(clicked()), SLOT(playStop()) );
  fwdStep= new QPushButton(this);
  fwdStep->setPixmap(QPixmap(fwd1icon));
  fwdStep->resize(fwdStep->sizeHint());
  connect(fwdStep, SIGNAL(clicked()), SLOT(windFwd1()) );
  fwdPlay= new QPushButton(this);
  fwdPlay->setPixmap(QPixmap(fwdcicon));
  fwdPlay->resize(fwdPlay->sizeHint());
  connect(fwdPlay, SIGNAL(clicked()), SLOT(playStart()) );
  fwdMax= new QPushButton(this);
  fwdMax->setPixmap(QPixmap(fwd0icon));
  fwdMax->resize(fwdMax->sizeHint());
  connect(fwdMax, SIGNAL(clicked()), SLOT(windFwdMax()) );

  destTime= new QLineEdit(this);
  destTime->setText("0");
  destTime->resize(destTime->sizeHint());
  destTime->setValidator(new QIntValidator( destTime ));

  playTo= new QPushButton(this);
  playTo->setPixmap(QPixmap(fwdtoicon));
  playTo->resize(playTo->sizeHint());
  connect(playTo, SIGNAL(clicked()), SLOT(playMaxTime()) );
  jumpTo= new QPushButton(this);
  jumpTo->setPixmap(QPixmap(jmpicon));
  jumpTo->resize(jumpTo->sizeHint());
  connect(jumpTo, SIGNAL(clicked()), SLOT(wind()) );

  btnfwddelay= new QPushButton(this);
  btnfwddelay->setPixmap(QPixmap(fwddelay));
  btnfwddelay->resize(btnfwddelay->sizeHint());
  connect(btnfwddelay, SIGNAL(clicked()), SLOT(setPlayDelay()));
  btnfwddelayi= new QPushButton(this);
  btnfwddelayi->setPixmap(QPixmap(fwddelayi));
  btnfwddelayi->resize(btnfwddelayi->sizeHint());
  connect(btnfwddelayi, SIGNAL(clicked()), SLOT(incPlayDelay()));
  btnfwddelayd= new QPushButton(this);
  btnfwddelayd->setPixmap(QPixmap(fwddelayd));
  btnfwddelayd->resize(btnfwddelayd->sizeHint());
  connect(btnfwddelayd, SIGNAL(clicked()), SLOT(decPlayDelay()));

  // initialize color select buttons ...
  int nCSel( FColors::numAgentColors() );
  cSel= new (FColorSelect *)[nCSel];

  for (int i=0; i<nCSel; i++) 
    cSel[i]= new FColorSelect(i, this);
}

void MainWidget::connectFieldMap( const FieldMap *fm )
{
#ifdef QT2
  QObject::connect( fm, SIGNAL( statusMsg ( const QString & ) ),
		    statusBar, SLOT(message( const QString &)));
#else
  QObject::connect( this, SIGNAL(statusMsg(const char *)),
		    statusBar, SLOT(message(const char *)));
#endif

  connect( fm, SIGNAL( markerSet ( Vector2d, int )),
	   SLOT( markerSet ( Vector2d, int )));
  connect( fm, SIGNAL( zoomWindow ( Vector2d, Vector2d, double)),
	   SLOT(  zoomWindow ( Vector2d, Vector2d, double)) );
  connect( fm, SIGNAL( sigFreekick ( Vector2d, bool )),
	   SLOT( sigFreekick ( Vector2d, bool )) );
  connect( fm, SIGNAL( sigDropBall ( Vector2d )),
	   SLOT(sigDropBall ( Vector2d )));
  connect( fm, SIGNAL( sigSetPlayer ( Vector2d )),
	   SLOT( sigSetPlayer ( Vector2d )));
}

void MainWidget::setMagSteps(FieldMap *fm, bool isMainMap)
{
  if (isMainMap) 
    fm->dflMagSteps(Config::getValueAsDouble("FMonDflPlayerMag"),
		    Config::getValueAsDouble("FMonDflBallMag"),
		    Config::getValueAsDouble("FMonDflSpeedMag"));
  else 
    fm->dflMagSteps(Config::getValueAsDouble("FMonZDflPlayerMag"),
		    Config::getValueAsDouble("FMonZDflBallMag"),
		    Config::getValueAsDouble("FMonZDflSpeedMag"));
}

list<FieldObject *> *MainWidget::basicList()
{
  list<FieldObject *> *l= new list<FieldObject *>;
  
  // NEVER add any FieldBackground here

  // make a copy of all FieldMarkers!
  const FieldMarker *m(markerwin->getFirstMarker(curTime()));
  // tmp holds the copy, if the marker exists
  FieldMarker *tmp;

  while (m) {
    tmp=new FieldMarker(*m);
    l->push_back(tmp);
    m= markerwin->getNextMarker();
  }
  return l;
}

void MainWidget::deleteList( list<FieldObject *> * &l)
{
  // check if there is a list
  if (l == 0)
    return;
  for( list<FieldObject *>::iterator it= l->begin();
       it != l->end(); it++)
    delete (*it);
  // this cleans up pointers using a clear():
  delete l;
  l = 0;
}
