//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef MSGSTORE_H
#define MSGSTORE_H
#include "SoccerMessage.h"
#include "Exception.h"
#include "ICFile.h"

class MsgStore {
public:
  enum {
    Buffered = 32,
    MaxUnstoredBytes = 100000
  };
  MsgStore(const string &file) throw (Exception);
  ~MsgStore();

  void openWrite() throw (Exception);

  bool insertMsg(const SoccerMessage &msg) throw (Exception);
  
  bool fetchTime(int time) throw (Exception);
  const SoccerMessage &getNextMessage() throw (Exception);

  int getMaximalSeenTime()  const { return writeTime; }
  int getCurrentFetchTime() const { return curTime;  }

  void show();

private:
  struct MsgItem {
    SoccerMessage   msg;
    struct MsgItem *next;
  };

  ICFile file;

  ////// data for writing
  // non-flushed messages stored as msg.time % StoredTimeSteps
  MsgItem *messages[Buffered];
  // the highest time written, if valid, i.e. >= -1
  int writeTime;
  int minwriteTime;
  int numUnstored;

  ////// file data
  // location in file
  struct Location {
    int fpos;
    int length;
  };
  vector<Location> location;
  int lastWrittenTime;
  Location writeLocation;

  ////// data for reading
  // if data is already written to disk, current timesteps msgs are here
  MsgItem *readmsg;
  // time that is currently read
  int curTime;
  // next message to read if readTime is valid
  MsgItem *curMsg;  
  // contains real location of cur_msg (list start or item)
  MsgItem **curMsgLoc;

  ///// methods
  int  numSaved() { return location.size(); };
  void checkNumUnstored();
  void setupLocations() throw (Exception);
  void addMsg(MsgItem *&, const SoccerMessage &msg);
  int  timeAddr(int t);
  void clearList(MsgItem *&);
  void appendToFile(int t) throw (Exception);
  void readFromFile(int t, MsgItem *&) throw (Exception);
};
#endif
