//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- /////////////////////////////////////////////////
//
// $Id: mastercontrol.C,v 1.75 2001/08/06 02:07:38 mrb Exp $
//
// 
// cwz-build-command: make mrbplayer
////////////////////////////////////////////////////////////////

#include "mastercontrol.h"

#include "drawMacros.h"
#include "Config.h"
#include "mcommand.h"
#include "AdvSkills.h"

#include <strstream.h>
#include <string>

#undef COUT
#undef ACTTRACE
#undef DRAWTRACE
#undef USETIMETRACE

// trace macro: uncomment the line you want.

//#define COUT(output) cout << getTime() << " MC: " << output << endl
//#define DRAWTRACE(output) cout << getTime() << "% " << output << endl
//#define ACTTRACE(output) cout << getTime() << " MC: " << output << endl
//#define USETIMETRACE
//#define USEMODSELECTTRACE
//#define USECOMMTRACE
//#define USETACTTRACE

//#define SHOWOFFSIDELINE

#ifdef USEMODSELECTTRACE
#define MODSELTRACE(cmd) cmd
#warning: MasterControl debug trace MODSELTRACE is activated!
#else
#define MODSELTRACE(output)
#endif

#ifdef USECOMMTRACE
#define COMMTRACE(cmd) cmd
#warning: MasterControl debug trace COMMTRACE is activated!
#else
#define COMMTRACE(output)
#endif

#ifdef USETACTTRACE
#define TACTTRACE(cmd) cmd
#warning: MasterControl debug trace TACTTRACE is activated!
#else
#define TACTTRACE(output)
#endif

#ifndef COUT
#define COUT(output)
#else
#warning: MC debug trace COUT is activated!
#endif

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning: MC debug trace DRAWTRACE is activated!
#endif

#ifndef ACTTRACE
#define ACTTRACE(output)
#else
#warning: MC debug trace ACTTRACE is activated!
#endif

#ifdef SHOWOFFSIDELINE
#warning: MC debug trace SHOWOFFSIDELINE is activated!
#endif

#ifndef USETIMETRACE
#define TIMETRACE(marker)
#define INITTTRACE
#define TOTALTTRACE
#else
#warning: MC debug trace TIMETRACE is activated!
#include <sys/resource.h>
#include <time.h>

static rusage start, total_start, end;

#define INITTTRACE \
getrusage(RUSAGE_SELF, &start); \
getrusage(RUSAGE_SELF, &total_start); \
cout << getTime() << " TT: starting" << endl; 

#define TIMETRACE(marker) \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: " marker " " << \
((1.0e-3*end.ru_utime.tv_usec   + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*start.ru_utime.tv_usec + 1.0e3*start.ru_utime.tv_sec)) << endl; \
start = end;

#define TOTALTTRACE \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: total time______________ " << \
((1.0e-3*end.ru_utime.tv_usec         + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*total_start.ru_utime.tv_usec + 1.0e3*total_start.ru_utime.tv_sec)) << endl;
#endif

MasterControl::MasterControl(const char* team, bool goalie, const char *host,
			     int port, bool log, int recn, 
			     const char *configfile, bool FUNSSELmode) :
  Player(team, goalie, host, port, log, recn, FUNSSELmode) ,
  lastTimeMsgSent(-1)
{
  if (!skills)
    exit(-666);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  goaleval     = new GoalShotEval();
  passeval     = new PassEval(goaleval);
  poseval      = new PositionEval(goaleval, passeval);

  tactical     = new TacticalModule();
  standard     = new StandardSituationModule(goalie, passeval, goaleval,
					     poseval);
  pass         = new PassModule(passeval, goaleval);
  ballhandling = new BallHandlingModule(goalie);
  goalshot     = new GoalShotModule(goalie, goaleval);
  position     = new PositionModule(goalie, passeval, goaleval, poseval);
}

MasterControl::~MasterControl() {

  delete tactical;
  delete standard;
  delete pass;
  delete passeval;
  delete ballhandling;
  delete goalshot;
  delete goaleval;
  delete position;
}

void
MasterControl::act() {
  INITTTRACE;

  skills->blockTurnHead();

  AudioMsg audiomsg;
  while (getNextMsg(audiomsg)) {
    COMMTRACE( { ostrstream ost;
    ost << "MC commtrace: got audiomsg: " << audiomsg << (char)0;
    Comm::sendMCmd(MLog(ost.str()));
    } );
    
    // first give all msgs to worldmodel, then modules
    worldFeedbackAudio(audiomsg);

    // only worldmodel can handle (shorter) coach msgs
    if (audiomsg.isCoachMsg()) {
        COMMTRACE( { ostrstream ost;
	ost << "MC commtrace: last audiomsg was coachmsg: " << audiomsg << 
	  (char)0;
	Comm::sendMCmd(MLog(ost.str()));
	} );	
      continue;
    }
    
    tactical->feedbackAudio(audiomsg);

    standard->feedbackAudio(audiomsg);
    pass->feedbackAudio(audiomsg);
    ballhandling->feedbackAudio(audiomsg);
    goalshot->feedbackAudio(audiomsg);
    position->feedbackAudio(audiomsg);
  }

  // get snapshot 
  
  Snapshot *snap = getSnapshot();
  
  if (snap == NULL) {
    cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	 <<"): MC: cannot generate Snapshot of time " << getTime() << endl;
    return; 
  } 
  
  // traces to check offside line
#ifdef SHOWOFFSIDELINE
  { 
    ostrstream ost;
    ost << "MC offside-line: ourOffsideLine: " 
	<< snap->ourOffsideLine(MAXAGE) 
	<< " otherOffsideLine: "
      	<< snap->otherOffsideLine(MAXAGE) 
	<< (char)0;
    Comm::sendMCmd(MLog(ost.str()));
    Comm::sendMCmd(MDVLine(snap->ourOffsideLine(MAXAGE)));
  } 
#endif

  COUT("doing world feedback for tactical module...");
  tactical->feedbackWorld(snap);

  // get tactical snapshot from tactical module

  TacticalSnapshot *tsnap = tactical->getSnapshot();

  TACTTRACE( { ostrstream os;
  os << "Tactical Snapshot (before eval): " << *tsnap << (char)0;
  Comm::sendMCmd(MLog( os.str() ));
  } )

  COUT("doing world feedback for standardsit module...");
  standard->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for pass module...");
  pass->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for ballhandling module...");
  ballhandling->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for goalshot module...");
  goalshot->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for position module...");
  position->feedbackWorld(snap, tsnap);

  TIMETRACE("after modules feed back ");

  // get all action evaluations from modules
  Grade stndG = standard->evaluate(skills, snap, tsnap);
  TIMETRACE("after standard eval     ");
  COUT("doing evaluate for standardsit module: " << stndG);
  Grade passG = pass->evaluate(skills, snap, tsnap);
  TIMETRACE("after pass     eval     ");
  COUT("doing evaluate for pass module: " << passG);
  Grade handG = ballhandling->evaluate(skills, snap, tsnap);
  TIMETRACE("after ballh    eval     ");
  COUT("doing evaluate for ballhandling module: " << handG);
  Grade goalG = goalshot->evaluate(skills, snap, tsnap);
  TIMETRACE("after goalshot eval     ");
  COUT("doing evaluate for goalshot module:" << goalG);
  Grade posnG = position->evaluate(skills, snap, tsnap);
  COUT("doing evaluate for position module:" << posnG );
  TIMETRACE("after position eval     ");

  enum { DEC_STAND, DEC_PASS, DEC_HAND, DEC_GOAL, DEC_POSN} decision= DEC_POSN;

  // simple bugfix: avoid undefined grades
  
  if (!stndG.is_defined()) {
    stndG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " stndG undefined! Changed to impossible"
	 << endl;
  }

  if (!goalG.is_defined()) {
    goalG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " goalG undefined! Changed to impossible"
	 << endl;
  }

  if (!handG.is_defined()) {
    handG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " handG undefined! Changed to impossible"
	 << endl;
  }

  if (!passG.is_defined()) {
    passG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " passG undefined! Changed to impossible"
	 << endl;
  }

  if (!posnG.is_defined()) {
    posnG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " posnG undefined! Changed to impossible"
	 << endl;
  }

  // handle absolute values
  
  if (stndG.is_ABSOLUTE()) 
    decision= DEC_STAND;
  else if (goalG.is_ABSOLUTE()) 
    decision= DEC_GOAL;
  else if (handG.is_ABSOLUTE())
    decision= DEC_HAND;
  else if (passG.is_ABSOLUTE())
    decision= DEC_PASS;
  else if (posnG.is_ABSOLUTE())
    decision= DEC_POSN;

  // Magic decision tree
  
  else if (!goalG.is_worse_than( handG ) &&
	   !goalG.is_worse_than( passG ) &&
	   !goalG.is_worse_than( posnG ) ) 
    decision= DEC_GOAL;
  else if (!passG.is_worse_than( handG ) && 
	   !passG.is_worse_than( posnG ) ) 
    decision= DEC_PASS;
  else if (!handG.is_worse_than( posnG ) &&
	   !handG.is_worse_than( passG ) ) 
    decision= DEC_HAND;
  else if (!passG.is_worse_than( posnG )) 
    decision= DEC_PASS;
  else 
    decision= DEC_POSN;
  
  // do actions

  ACTTRACE(snap->my.number<< ": ;std="<< stndG << ", goal=" << goalG << ", pass=" << passG 
	   << ", bhand=" << handG << ", pos=" << posnG );
  COUT("calling one module to act...");

  MODSELTRACE( { ostrstream ost;
	       ost << "MC Grade eval: std="<< stndG << ", goal=" 
	       << goalG << ", pass=" << passG 
	       << ", bhand=" << handG << ", pos=" << posnG << (char)0;
	       Comm::sendMCmd(MLog(ost.str()));
  } );

  switch (decision) {
  case DEC_STAND:
    ACTTRACE( "handling standard situation, grade=" << stndG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC act: std")););
    standard->act(skills, snap, tsnap);
    break;
  case DEC_PASS:
    ACTTRACE( "passing, grade=" << passG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC act: pass")););
    pass->act(skills, snap, tsnap);
    break;
  case DEC_HAND:
    ACTTRACE( "handling ball, grade=" << handG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC act: ballh")););
    ballhandling->act(skills, snap, tsnap);
    break;
  case DEC_GOAL:
    ACTTRACE( "kicking into goal, grade=" << goalG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC act: goal")););
    goalshot->act(skills, snap, tsnap);
    break;
    // default action: positioning ...
  case DEC_POSN:
  default:
    ACTTRACE( "positioning, grade=" << posnG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC act: posn")););
    position->act(skills, snap, tsnap);
  } // switch

  TIMETRACE("after module act        ");

  TACTTRACE( { ostrstream os;
  os << "Tactical Snapshot (before act): " << *tsnap << (char)0;
  Comm::sendMCmd(MLog( os.str() ));
  } )

  // now collect all TurnHeadCommands ...
  AdvSkills::TurnHeadCommand stdTHC, passTHC, handTHC, goalTHC, posTHC;

  // now do similar evaluation for turn head ...
  stndG = standard->turnHeadEvaluate(skills, stdTHC );
  TIMETRACE("after standard eval     ");
  COUT("doing evaluate for standardsit module: " << stndG);
  passG = pass->turnHeadEvaluate(skills, passTHC);
  TIMETRACE("after pass     eval     ");
  COUT("doing evaluate for pass module: " << passG);
  handG = ballhandling->turnHeadEvaluate(skills, handTHC);
  TIMETRACE("after ballh    eval     ");
  COUT("doing evaluate for ballhandling module: " << handG);
  goalG = goalshot->turnHeadEvaluate(skills, goalTHC);
  TIMETRACE("after goalshot eval     ");
  COUT("doing evaluate for goalshot module:" << goalG);
  posnG = position->turnHeadEvaluate(skills, posTHC);
  COUT("doing evaluate for position module:" << posnG );
  TIMETRACE("after position eval     ");

  MODSELTRACE( { ostrstream ost;
	       ost << "MC TurnHeadGrade eval: std="<< stndG << ", goal=" 
	       << goalG << ", pass=" << passG 
	       << ", bhand=" << handG << ", pos=" << posnG << (char)0;
	       Comm::sendMCmd(MLog(ost.str()));
  } );

  // simple bugfix: avoid undefined grades
  
  if (!stndG.is_defined()) {
    stndG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " TH stndG undefined! Changed to impossible"
	 << endl;
  }

  if (!goalG.is_defined()) {
    goalG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " TH goalG undefined! Changed to impossible"
	 << endl;
  }

  if (!handG.is_defined()) {
    handG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " TH handG undefined! Changed to impossible"
	 << endl;
  }

  if (!passG.is_defined()) {
    passG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " TH passG undefined! Changed to impossible"
	 << endl;
  }

  if (!posnG.is_defined()) {
    posnG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << snap->my.number << " TH posnG undefined! Changed to impossible"
	 << endl;
  }


  /*********************************************************
   ********************************************************* 
   * CHECK THIS DECISION TREE AND DISCUSS DEFAULT ACTION ! *
   *********************************************************
   *********************************************************/

  // now decide which module gets the turn head ...
  // (another magic decision tree)

  enum { TH_NONE, TH_STAND, TH_PASS, TH_HAND, TH_GOAL, TH_POSN} thdecision= 
								  TH_NONE;

  if (stndG.is_ABSOLUTE()) 
    thdecision= TH_STAND;
  else if (handG.is_ABSOLUTE())
    thdecision= TH_HAND;
  else if (goalG.is_ABSOLUTE()) 
    thdecision= TH_GOAL;
  else if (passG.is_ABSOLUTE())
    thdecision= TH_PASS;
  else if (posnG.is_ABSOLUTE())
    thdecision= TH_POSN;

  // Magic decision tree
  
  else if (!handG.is_worse_than( goalG ) &&
	   !handG.is_worse_than( passG ) &&
	   !handG.is_worse_than( posnG ) &&
	   !handG.is_IMPOSSIBLE() ) 
    thdecision= TH_HAND;
  else if (!goalG.is_worse_than( passG ) &&
	   !goalG.is_worse_than( posnG ) &&
	   !goalG.is_IMPOSSIBLE()) 
    thdecision= TH_GOAL;
  else if (!passG.is_worse_than( posnG ) &&
	   !passG.is_IMPOSSIBLE() ) 
    thdecision= TH_PASS;
  else if (!posnG.is_IMPOSSIBLE())
    thdecision= TH_POSN;

  switch (thdecision) {
  case TH_STAND:
    ACTTRACE( "MC turnHead: std" );
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: std")););
    
    skills->exec(snap, stdTHC);

    break;
  case TH_PASS:
    ACTTRACE( "MC turnHead: pass" );
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: pass")););

    skills->exec(snap, passTHC);

    break;
  case TH_HAND:
    ACTTRACE( "MC turnHead: ballh" );
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: ballh")););

    skills->exec(snap, handTHC);

    break;
  case TH_GOAL:
    ACTTRACE( "kicking into goal, grade=" << goalG );
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: goal")););

    skills->exec(snap, goalTHC);

    break;
  case TH_POSN:
    ACTTRACE( "MC turnHead: posn" );
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: posn")););

    skills->exec(snap, posTHC);
    break;
  default:
    // default action: no turn head!
    MODSELTRACE(Comm::sendMCmd(MLog("MC turnHead: NONE")););
  } // switch


  // audio message generation
  COUT( "testing for message to be sent..." );
  switch (snap->getChannelNow()) {
    
  case AUMUL_NONE: // empty channel never used, to make sure emergency channel is free
    COUT( "no sending allowed before emergency channel" );
    break;
  
  case AUMUL_EMRG: { // if emergency channel is used, worldbroadcast will be blocked
    if (tactical->useEmergencyChannel) {
      sendAudioMsg(snap);
      tactical->useEmergencyChannel = false;
      COMMTRACE( { ostrstream ost;
      ost << "MC commtrace: Sending msg on emergency channel." << (char)0;
      Comm::sendMCmd(MLog(ost.str()));
      } );
    } else {
      COUT( "no sending on emergency channel" );
    }
  } break;
    
  case AUMUL_MY_WM:  // broadcast channel always used, but only "blocks" empty channel
    // using lastTimeMsgSent to avoid saying two msgs in a 3-cycle-slice
    if (snap->time - lastTimeMsgSent > 2) {
      sendAudioMsg(snap);
      COMMTRACE( { ostrstream ost;
      ost << "MC commtrace: Sending msg on worldbroadcast." << (char)0;
      Comm::sendMCmd(MLog(ost.str()));
      } );
    } else {
      COUT( "no worldmodel broadcast because of other msg in:" << lastTimeMsgSent );
    }
    break;
  case AUMUL_OTHER_WM:
    COUT( "not my turn sending worldmodel broadcast" );
    break;
    
  default:;
  } // switch

  COMMTRACE( { ostrstream ost;
  int ntime(snap->nextTimeOfChannel(AUMUL_NONE));
  int etime(snap->nextTimeOfChannel(AUMUL_EMRG));
  int mwtime(snap->nextTimeOfChannel(AUMUL_MY_WM)); 
  int owtime(snap->nextTimeOfChannel(AUMUL_OTHER_WM));
  ost << "MC commtrace: statistics: (n,e,m,o)= (" << ntime << ", " << etime 
      << ", " << mwtime << ", " << owtime << ")" << (char)0;
  Comm::sendMCmd(MLog(ost.str()));
  } );

  TOTALTTRACE;
} //act


void MasterControl::sendAudioMsg(const Snapshot *snap)
{
  if (snap->time == lastTimeMsgSent) {
    COUT( "Already sent message in this cycle");
  }  else {
    lastTimeMsgSent= snap->time;
    
    AudioMsg sendMsg(0, snap->time);
  
    bool msgState= true;
  
    worldGetSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();
  
    tactical->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    standard->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();
  
    pass->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    ballhandling->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    goalshot->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    position->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    if (msgState) {
      say(sendMsg);
      COUT("Sending AudioMsg: " << sendMsg );
    } else {
      cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	   <<"): MC: Fatal error: buffer overflow: "<< sendMsg
	   << getTime() << endl;
    }   
  }
}

  
