#include <iostream.h>
#include "Perm.h"

/////////////////////////////////////////////////////////////////////
//
//     class Permutation
//
////////////////////////////////////////////////////////////////////

Permutation::
Permutation( const Permutation &perm ) : p(NULL), n(perm.n)
{
  p = new char[n];
  for (char i = 0; i < n; i++) 
    p[i] = perm[i];
}

inline
char
Permutation::
operator[] ( const char pos ) const
{
  if (p && (0 <= pos) && (pos < n))
    return p[pos];
  else
    return -1;
}

Permutation&
Permutation::
operator= ( Permutation &perm )
{
  delete p;
  n = perm.n;
  p = new char[n];

  for (char i = 0; i < n; i++)
    p[i] = perm[i];

  return perm;
}

Permutation::
Permutation( const char _n ) : p(NULL), n(_n)
{
  if (n > 0) 
    p = new char[n];
}

char*
Permutation::
operator= ( char* perm )
{
  for (char i = 0; i < n; i++) 
    p[i] = perm[i];

  return perm;
}

/////////////////////////////////////////////////////////////////////
//
//     class Permutations
//
////////////////////////////////////////////////////////////////////

Permutations::
Permutations( const char _n, const bool gen ) 
  : n(_n), fac(0), generated(false), p(NULL)
{
  fac = nfac(n);
  p = new Permutation[fac];

  Permutation P(n);
  for (unsigned int i = 0; i < fac; i++)
    *(p+i) = P;

  if (gen) generate();
}

Permutations::
Permutations( const Permutations &perm )
  : n(perm.n), fac(perm.fac), generated(perm.generated), p(NULL)
{
  p = new Permutation[fac];
  if (generated)
    for (unsigned int i = 0; i < fac; i++)
      *(p+i) = perm[i];
  else {
    Permutation P(n);
    for (unsigned int i = 0; i < fac; i++)
      *(p+i) = P;
  }
}

Permutations
Permutations::
of( const char _n )
{
  char N(0);
  if (abs(_n) > n)
    N = n;
  else
    N = abs(_n);

  if (!generated)
    generate();

  Permutations P(N, false);

  for (unsigned int i = 0; i < P.fac; i++)
    for (char j = 0; j < N; j++)
      *((*(P.p+i)).p+j) = p[fac-P.fac+i][n-N+j];

  P.generated = true;

  return P;  
}

inline
Permutation&
Permutations::
operator[] ( const unsigned int pos ) const
{
  if ((0 <= pos) && (pos < fac) )
    return *(p+pos);
  else
    return *(p+0);
}


inline
unsigned int
Permutations::
nfac( const int n ) const
{
  unsigned int result(1);
  int i(2);
  while (i <= n) result *= i++;
  return result;
}

void
Permutations::
generate( void )
{
  if (!generated) {
    unsigned int l(0);
    char *perm = new char[n];
    for (char i = 0; i < n; i++)
      perm[i] = -1;
    genPerm(perm,0,l);
    delete perm;
    generated = true;
  }
}

void 
Permutations::
genPerm(char perm[], char k, unsigned int &l) {

  if (k == n) {
    *(p+l) = perm;
    l++;
    return;
  }

  for (char i = 0; i < n; i++) {
    bool used(false);
  
    for (char j = 0; j < k; j++)
      if (perm[j] == i) {
	used = true; 
	break;
      }

    if (!used) {
      perm[k] = i;

      genPerm(perm,k+1,l);
      perm[k] = -1;
    }
  }
}
