//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-////////////////////////////////////////////////////////
//
// $Id: coach.C,v 1.15 2001/06/04 14:26:01 satan Exp $
//
// coach.C - interface for training coach client
//
///////////////////////////////////////////////////////////////////////

#include "coach.h"

#undef TRACE
#define TRACE(cmd)


Coach::Coach (const char*  team_name,             
	      playing_side side,
	      const char*  host,
	      int          port,
	      bool         logging_snap,
	      bool         FUNSSELmode,
	      int          numSnapshots) :
  
  comm(NULL),
  world(NULL)
{
  TRACE (cout << "Coach: reading constants" << endl);
    
  TRACE (cout << "Coach: generating Comm" << endl);
    
  comm = new Comm(team_name, FUNSSELmode);

  TRACE (cout << "Coach: initializing Comm" << endl);
    
  comm_error ce;
  ce = comm->init(CLIENT_TRAINCOACH, host, port, 0, logging_snap, side);
    
  switch (ce) {
  case (COMM_OK) : 
      
    TRACE (cout << "Coach: generating Worldmodel" << endl);
    
    world = new Worldmodel (numSnapshots, comm);

    writeSnaps = false;
    if (logging_snap) {
      writeSnaps = true;
      // open snap_log file
      ostrstream tmp;
      tmp << SNAP_LOG_FILENAME << getMyTeam() << getMyNumber() 
	  << SNAP_LOG_EXTENSION;
      char* filename = tmp.str();
      out.open(filename, ios::out);
      // Version control
      out << SNAPLOG_PROTOCOL << endl;
      delete filename;
    }
    break;
  case (COMM_ERROR): 
    cerr << "Coach: unspecified error when initializing !" << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Coach: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Coach: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Coach: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_SIGNAL_INIT):
    cerr << "Coach: could not set up signal handlers !" << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Coach: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Coach: did not receive response from server !" << endl; break;
  }
  
}


Coach::~Coach () {

  TRACE (cout << "Coach: entering destructor, closing down" << endl);

  delete (world);
  delete (comm);

  if (writeSnaps) out.close();

  TRACE (cout << "Coach: leaving destructor" << endl);

}


void 
Coach::coach() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Coach: started" << endl);

    while (comm->isServerAlive() && !comm->isGameOver() ) {
      
      TRACE (cout << "Coach: waiting for signals and handling them" << endl);
      comm->handleSignals();

      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Coach: adjusting worldmodel" << endl);
	world->adjust();
      
	TRACE (cout << "Coach: getting snapshot of time " 
	       << getTime() << endl);
	Snapshot* snap = world->getSnapshot(getTime());
	
	if (snap == NULL) {
	  cerr << "Coach (" << getMyTeam()
	       <<"): cannot generate Snapshot of time " << getTime() << endl;
	} else {
	  
	  if ( writeSnaps && comm->isTimeRunning() ) {
	    snap->binaryWriteToStreamOSI(out);
	    out << endl;
	  }
	  
	  TRACE (cout << "Coach: acting" << endl);
	  act(snap);
	
	}
	
      } else {
	TRACE(cout << "Coach: no new timestep yet" << endl);
      }

    } // while server alive

    TRACE (cout << "Coach: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Coach: could not connect to sserver, closing down" << endl;
  }
  
} //coach


void
Coach::changeMode(play_mode pm) {
  comm->sendChangeMode(pm);
}

void
Coach::changePlayerType(int player, int type) {
  comm->sendChangePlayer(player, type);
}

void
Coach::ear(bool earOn) {
  comm->sendEar(earOn);
}

void
Coach::moveObject (const char *object, double x, double y) {
  comm->sendMoveObject(object, x, y);
}

void
Coach::moveObject (const char *object, double x, double y, 
		   double dir_rad) {
  comm->sendMoveObject(object, x, y, dir_rad);
}

void
Coach::moveObject (const char *object, double x, double y, 
		   double dir_rad, double xvel, double yvel) {
  comm->sendMoveObject(object, x, y, dir_rad, xvel, yvel);
}

void
Coach::recover() {
  comm->sendRecover();
}
