//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module ballHandling
//
// $Id: ballhandlingmodule.C,v 1.78 2001/07/10 12:31:04 aarn Exp $
//
//////////////////////////////////////////////////////////////////////


#include "ballhandlingmodule.h"
#define FUNSSELMACROS
#include "drawMacros.h"
#include <strstream>

// definition of trace macros
////////////////////////////////////////////////////////////////

//#define BALLHTRACE

// general tracing, used to observe functioning of the module
#define CERR(output) cerr << time << " BALLH: " << output << endl

#ifdef BALLHTRACE
#warning: BALLH debug trace is activated!
#define BHTRC(output) Comm::sendMCmd(output)
#define GRADETRC(output, grade) { \
strstream tmp; \
tmp << output << " -> grade=" << grade << (char)0; \
Comm::sendMCmd(MLog(tmp.str())); \
}
#else
#define BHTRC(output)
#define GRADETRC(output, grade)
#endif

BallHandlingModule::BallHandlingModule(bool _goalie)
  : goalie(_goalie),
    sendSomething(false),
    sendInfo(new BallHandlingModuleInfo()),
    lastKickTime(-2),
    lastReportTime(-2),
    lookingForBall(false)
{
}

BallHandlingModule::~BallHandlingModule() {
  delete (sendInfo);
} 



/////////////////////////////////////////////////////////////////////////
//
// FEEDBACK
//
/////////////////////////////////////////////////////////////////////////



bool 
BallHandlingModule::feedbackWorld(const Snapshot *_snap, 
				  TacticalSnapshot *_tsnap) {

  // returns true if given data (which is to be used for evaluate) is valid
  // and could be stored temporarily,
  // false otherwise
  if ((_snap == NULL) || (_tsnap == NULL)) {
    if (time > -1)
      CERR("data update -> failed");
    return false;
  }

  snap = _snap;
  tsnap = _tsnap;
    
  playMode = snap->playMode;
  time     = snap->time;
  my       = snap->my;
  stamina  = snap->senseBody.stamina;
  ball     = snap->ball;
  ourGoal  = snap->ourGoal;
  distToBall       = my.distTo(ball);

  // Noda checks for <= stamina_limits, so rather stay slightly above:
  availableStamina = max (0.0, stamina - EFFORT_DEC_THR * STAMINA_MAX - 1 );

  faceAngleToBall  = my.faceAngleTo(ball);
  bodyAngleToBall  = my.bodyAngleTo(ball);

  if (snap->ball.age <= 1)
    lookingForBall = false;

  if ((tsnap->firstDangerousOpponent != UNKNOWN) &&
      (tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) <= 1)) {
    lastKickTime = time;
    opponentKicked = true;
  }

  if ((tsnap->firstResponsiblePlayer != UNKNOWN) &&
      (tsnap->ourTeamReachBallTime(tsnap->firstResponsiblePlayer) <= 1)) {
    // note that this also fits for the player itself
    // so we should look after our own kicks
    lastKickTime = time;
    opponentKicked = false;
  }

  return true;
}

bool 
BallHandlingModule::feedbackAudio(AudioMsg &audioMsg) {
  
  bool hasInfo;

  // read info from stream
  audioMsg >> hasInfo;
  
  if (hasInfo) {
  
    BallHandlingModuleInfo receivedInfo;
    audioMsg >> receivedInfo;

    // evaluate received information into module member variables here
    // (not used yet)
    
  }

  return true;
}

/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingEvaluate.C"

/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingAct.C"

///////////////////////////////////////////////////////////////////////////
//
// SAY MSG and Info Class
//
///////////////////////////////////////////////////////////////////////////

bool 
BallHandlingModule::getSayMsg(AudioMsg &audioMsg) {
  
  if (sendSomething) {
    // will currently never be true, communication between ballhandlingmodules
    // is still unused.

    audioMsg.markWrite();
    
    audioMsg << true;

    audioMsg << sendInfo;
    
    // check if error occurred
    if( !audioMsg.getState() )
      {
	cerr << "BALLH: Stream overflow: " << audioMsg << endl;
	audioMsg.rollbackWrite().clearState();

	audioMsg << false;

      }
  } else {

    audioMsg << false;

  }
  
  return true;
  
}


BallHandlingModuleInfo::BallHandlingModuleInfo()
{
}

BallHandlingModuleInfo::~BallHandlingModuleInfo()
{
}

class AudioMsg &operator<< (AudioMsg &aumsg, 
			    const BallHandlingModuleInfo &) {
  // write all data from Info class here, and make sure there are 
  // AudioMsg-operators defined for the specific type of data you write !
  
  // for example:
  // aumsg << info.value1;
  // aumsg << info.value2;

  return aumsg;
}


class AudioMsg &operator>> (AudioMsg &aumsg, 
			    BallHandlingModuleInfo &) {
  // read data for Info class in same order as writing it

  // aumsg >> info.value1;
  // aumsg >> info.value2;
  
  return aumsg;
}
