//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module GoalShotModule
//
// \$Id: goalshotmodule.C,v 1.16 2001/08/06 02:22:39 satan Exp $
// \$manager: $schappel
//
//////////////////////////////////////////////////////////////////////

#include "goalshotmodule.h"

// the GOALTRACE
////////////////////////

#undef USEGOALTRACE


// use this line if you want prompting of players decision process to FUNSSEL log
//#define USEGOALTRACE


#ifdef USEGOALTRACE
#define GOALTRACE(cmd) cmd
#warning GoalShot debug trace GOALTRACE is activated!
#else
#define GOALTRACE(cmd)
#endif


////////////////////////////
//
// Constructor
//
///////////////////////////

GoalShotModule::GoalShotModule(bool isGoalie, GoalShotEval *_goalshoteval):
  goalie(isGoalie),
  goalshoteval(_goalshoteval),
  sendInfo(NULL),
  // [C]
  mustRekick(false),
  // [C]
  finishedKick(false),
  lastSnapshotTime(-1),       // module has never seen a snaphshot
  availableKickTime(0),
  goalTarget(Vector2d(FIELD_LENGTH /2, 0)),
  // [C]
  useSuperKick(false),
  // [C]
  useFixedKick(false)
{
  if(goalshoteval)
    maxAge = goalshoteval->maxAge;
  else
    maxAge = 20;
};

////////////////////////////
//
// Destructor
//
///////////////////////////

GoalShotModule::~GoalShotModule()
{
};


/////////////////////////////////////////////////////////////////////
//
// public methods
//
/////////////////////////////////////////////////////////////////////

//////////////////////////////////////////
//
// bool GoalShotModule::feedbackWorld
//
/////////////////////////////////////////

bool GoalShotModule::feedbackWorld(const Snapshot* _worldSnap, 
				   TacticalSnapshot* _tacticalSnap)

{
  worldSnap = _worldSnap;
  tacticalSnap = _tacticalSnap;
  
  // process new snapshot 
  updateDataStorage();

  // must the player report a goalkick?
  if(!tacticalSnap)
    cerr << "GOAL: feedbackWorld -> tacticalSnap not valid" ;
  else
    if(finishedKick)
      {
	GOALTRACE({ostrstream ost;
	ost << "GOAL: feedbackWorld-> there is a finished goal kick"
	    << char(0);
	Comm::sendMCmd(MLog(ost.str()));
	});
	
	finishedKick=false;
	// did the player get a see-Information about the ball?
	if(worldSnap->ball.age == 1)
	  {
	    // will the ball stop behind the goalline?
	    if( (worldSnap->ball.pos 
		 + worldSnap->ball.v *(1/(1-BALL_DECAY))).getX() 
		>= FIELD_LENGTH /2.)
	      {
		// calculate the parameter t, when the line from ball.pos 
		// to ball.pos + ball.v*(1/(1-BALL_DECAY))
		// intersects the goalline
		double t = FIELD_LENGTH /2. - worldSnap->ball.pos.getX();
		t /= worldSnap->ball.v.getX();
		// does the ball pass the goalline within the goal?
		if (fabs(worldSnap->ball.pos.getY() 
			 + worldSnap->ball.v.getY() * t ) 
		    < GOAL_WIDTH / 2.) 
		  // report a goalkick
		  {
		    tacticalSnap ->reportOurTacticalAction(TA_GOALSHOT, 
							   me.number);
		    GOALTRACE({ostrstream ost;
		    ost << "GOAL: feedbackWorld -> goalshot reported"
			<< char(0);
		    Comm::sendMCmd(MLog(ost.str()));
		    });
		  }
	      }
	  }
		
      }
  
  return true;
}; 
  
//////////////////////////////////////////
//
// bool GoalShotModule::feedbackAudio
//
/////////////////////////////////////////

bool GoalShotModule::feedbackAudio(AudioMsg &audioMsg)

{
  GoalShotModuleInfo receivedInfo(false);
  audioMsg >> receivedInfo;
  if (receivedInfo.valid)
    {
      GOALTRACE({ostrstream ost;
      ost << "GOAL: got valid info"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      // evaluate
    }
  else
    {
      GOALTRACE({ostrstream ost;
      ost << "GOAL: got invalid info"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }

  return true;
}; 

//////////////////////////////////////
//
// Grade GoalShotModule::evaluate
//
/////////////////////////////////////

Grade GoalShotModule::evaluate(AdvSkills* skills, 
			       const Snapshot* _worldSnap, 
			       TacticalSnapshot* )
{
  Grade grade;

  delete (sendInfo);
  sendInfo = new GoalShotModuleInfo(false);
  
  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: act -> snapshot is NULL";
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // are the skills available?
  if(skills == NULL)
    {
      cerr << worldSnap->time << "GOAL: eval -> no skills available";
      grade.set_IMPOSSIBLE();
      return grade;
    }
 

  // ball not in the kick range
  /////////////////////////////////
  if (!skills->kickSafe(worldSnap)) 
    {
      grade.set_IMPOSSIBLE();
      mustRekick = false;
      GOALTRACE ({ostrstream ost;
      ost << "GOAL: eval -> ball not kickable; grade = " << grade 
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      return grade;
    }

  // do we need a rekick?
  ///////////////////////////////
  if (mustRekick)
    {
      grade.set_ABSOLUTE();
      GOALTRACE ({ostrstream ost;
      ost << "GOAL: eval -> rekick must be done; grade = "
	  << grade << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      return grade;
    }

  // now we are able to kick the ball 
  ////////////////////////////////////////

  // calculate the minimum of the players distance to the left edge
  // the middle or the right edge of the goal for calculating a first
  // goal target

  Vector2d leftTarget;
  Vector2d middleTarget;
  Vector2d rightTarget;
  double dummydist;
  goalshoteval->targetInGoal(me.pos, goalTarget, leftTarget, 
			     middleTarget, rightTarget, dummydist);


  // how much time does the player have for kicking ?
  availableKickTime = goalshoteval->availableTimeForKick(worldSnap, me.pos);
  GOALTRACE({ostrstream ost;
  ost << "GOAL: eval -> available time for kick = " << availableKickTime 
      << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  });

  // searching for the best corridor and evaluating it
  grade = searchAndEvalBestCorridor(skills, goalTarget, 
				    leftTarget, middleTarget,
				    rightTarget,
				    availableKickTime);

  return grade;
};

///////////////////////////////////
//
// GoalShotModule::turnHeadEvaluate
//
///////////////////////////////////

Grade GoalShotModule::turnHeadEvaluate(AdvSkills* skills, 
				       AdvSkills::TurnHeadCommand &thc)
{
  Grade g;

  WorldPlayer goalie;

  if (worldSnap->otherGoalieNo != UNKNOWN)
    {
      goalie = worldSnap->otherTeam(worldSnap->otherGoalieNo);
      /* if goalie is old and ball is not too old look at it, if
	 ball is close to me and I am close to their goal
       */
	if (goalie.age > 3 &&
	    dist(me.pos, worldSnap->ball.pos) < 2*VISIBLE_DISTANCE &&
	    worldSnap->ball.age < 3 &&
	    me.pos.getX() > FIELD_LENGTH/4.)
	  {
	    g.set_PERFECT();

	    // if goalie is not too old look directly at it
	    if (goalie.age < maxAge - 5)
	      {
		thc = skills->turnHead(worldSnap, me.bodyAngleTo(goalie.pos)+
				       me.bodyDir);
		thc.vwidth = VIEW_WIDTH_NARROW;
		GOALTRACE ({ostrstream ost;
		ost << "GOAL: turnHeadEvaluate -> want to turn head with narrow to other goalie." 
		    << " returning " << g
		    << char(0);
		Comm::sendMCmd(MLog(ost.str()));
		});
	      }
	    else // look at the middle of the goal
	      {
		thc = skills->turnHead(worldSnap, me.bodyAngleTo(worldSnap->otherGoal)+
				       me.bodyDir);
		GOALTRACE ({ostrstream ost;
		ost << "GOAL: turnHeadEvaluate -> want to turn head to other goal." 
		    << " returning " << g
		    << char(0);
		Comm::sendMCmd(MLog(ost.str()));
		});
	      }
	    return g;
	  }
    }

  g.set_IMPOSSIBLE();
  return g;
}


////////////////////////////
//
// GoalShotModule::act
//
///////////////////////////

bool GoalShotModule::act(AdvSkills* skills,
			 const Snapshot* worldSnap, 
			 TacticalSnapshot* tacticalSnap)
{
  // are the skills available?
  if(skills == NULL)
    {
      cerr << "GOAL: act -> no skills available" ;
      return false;
    }

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: act -> snapshot is NULL" ;
      skills->noAction();
      return true;
    }

  
  // the act-routine was called also the ball was not kickable,
  // so do nothing
  // [C]
  if(!skills->kickSafe(worldSnap))
    {
      GOALTRACE ({ostrstream ost;
      ost << "GOAL:  act -> ball not kickable, send no Action" 
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      mustRekick = false;
      skills->noAction();
      return true;
    }

  if(!tacticalSnap)
    // tactical snapshot is not valid => error message
    cerr << "GOAL: act -> tactical snapshot is NULL!" ;

  // [C]
  // player started a superkick and has not finished it yet
  if(mustRekick)
    {
      GOALTRACE ({ostrstream ost;
      ost << "GOAL: act -> rekicking" << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      kickParams = skills->rekick(worldSnap, kickParams);
      skills->exec(worldSnap, kickParams);
      if (kickParams.state == AdvSkills::KFINISHING)
	// superkick is finished after this kick
	{
	  mustRekick = false;
	  finishedKick = true;
	  GOALTRACE({ostrstream ost;
	  ost << "GOAL: act -> mustRekick has finished" 
	      << char(0);
	  Comm::sendMCmd(MLog(ost.str()));
	  });
	}
      return true;
    }

  bool reallyKicked = true;

  // [C] Eventuell anpassen, fasll skills angepasst werden

  if(dist(me.pos, goalTarget) < GOAL_WIDTH /2)
    // do not use too much time for a kick
    availableKickTime = min( availableKickTime, 2 );
  
  // if ball can be very fast with just 2 kicks, do not perform a superkick
  if(availableKickTime > 2)
    {
      kickParams = skills->kick(worldSnap, BALL_SPEED_MAX, 
				goalTarget, 2);
      if(kickParams.v.getLength() >= 2.0)
	{
	  GOALTRACE ({ostrstream ost;
	  ost << "GOAL: act -> no superkick; ball is fast enough" 
	      << char(0);
	  Comm::sendMCmd(MLog(ost.str()));
	  });
	  availableKickTime = 2;
	}
    }
  
  GOALTRACE ({ostrstream ost;
  ost << "GOAL: act -> kick with maxTime " << availableKickTime 
      << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  });
  kickParams = skills->kick( worldSnap, BALL_SPEED_MAX, 
			     goalTarget,
			     availableKickTime);
  skills->exec(worldSnap, kickParams);

  // [C] noch noetig nach Anpassung der skills??

  if(reallyKicked)
    // player sends a kick command
    if(kickParams.state != AdvSkills::KFINISHING)
      {
	mustRekick = true;
	finishedKick = false;
      }
    else
      {
	finishedKick = true;
	GOALTRACE({ostrstream ost;
	ost <<"act -> kick finished" 
	    << char(0);
	Comm::sendMCmd(MLog(ost.str()));
	});
      }

  return true; 
};  


/////////////////////////////////
//
// GoalShotModule::getSayMsg
//
////////////////////////////////


bool GoalShotModule::getSayMsg(AudioMsg &audioMsg)
{
  audioMsg.markWrite();
  audioMsg << sendInfo;

  // check if error occurred
  if( !audioMsg.getState() )
    {
      cerr << "GOAL: Stream overflow: " << audioMsg ;
      audioMsg.rollbackWrite().clearState();
      
      audioMsg << false;
      
    }

  return true;
};


////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////

/////////////////////////////////////////
//
// GoalShotModule::updateDataStorage
//
////////////////////////////////////////

bool GoalShotModule::updateDataStorage()
{

  // the player himself
  me = worldSnap->my;

  return true;
}

///////////////////////////////////////////
//
// bool GoalShotModule::targetReachable
//
//////////////////////////////////////////

bool GoalShotModule::targetReachable(AdvSkills *skills, 
				     Vector2d target, int availableTime)
{
  // [C] introduced that for not to use skills (CM)

  if( dist( worldSnap->ball.pos, target ) > 30 )
    return false;

  if( dist( worldSnap->ball.pos, target ) > 20 && availableTime <= 1 )
    return false;

  return true;
}



///////////////////////////////////////////////
//
// GoalShotModule::searchAndEvalBestCorridor
//
//////////////////////////////////////////////

Grade GoalShotModule::searchAndEvalBestCorridor(AdvSkills *skills, 
						Vector2d &target,
						Vector2d leftTarget,
						Vector2d middleTarget,
						Vector2d rightTarget,
						int availableTime)
{
  Grade grade;

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors(s) are possible by the given 
  // availableTime for a kick

  bool useLeftCorridor   = targetReachable( skills, leftTarget, availableTime );
  bool useMiddleCorridor = targetReachable( skills, middleTarget, availableTime );
  bool useRightCorridor  = targetReachable( skills, rightTarget, availableTime );

  if(!useLeftCorridor )
    {
      GOALTRACE({ostrstream ost;
      ost << "GOAL: searchAndEval->left Edge not reachable; left corridor unusable"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }
  if(!useMiddleCorridor )
    {
      GOALTRACE({ostrstream ost;
      ost << "GOAL: searchAndEval->middle of goal not reachable; middle corridor unusable"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }
  if(!useRightCorridor)
    {
      GOALTRACE({ostrstream ost;
      ost << "GOAL: searchAndEval->right edge not reachable; right corridor unusable"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }

  grade = goalshoteval->baseSearchAndEvalForCorridor(worldSnap, 
						     me.pos,
						     target, 
						     leftTarget,
						     middleTarget,
						     rightTarget,
						     useLeftCorridor,
						     useMiddleCorridor,
						     useRightCorridor);
  return grade;
}



/////////////////////////////////////////////
//
// GoalShotModuleInfo::GoalShotModuleInfo
//
/////////////////////////////////////////////

GoalShotModuleInfo::GoalShotModuleInfo(bool _valid)
{
  valid = _valid;
}

/////////////////////////////////////////////
//
// GoalShotModuleInfo::~GoalShotModuleInfo
//
/////////////////////////////////////////////

GoalShotModuleInfo::~GoalShotModuleInfo()
{
}


/////////////////////////////////////////////
//
// AudioMsg &operator<<
//
/////////////////////////////////////////////

class AudioMsg &operator<< (AudioMsg &aumsg, const GoalShotModuleInfo &info)
{
  aumsg << info.valid;
  if (info.valid);
  // aumsg << //further data structures to write ;"
  return aumsg;
}


/////////////////////////////////////////////
//
// AudioMsg &operator>>
//
/////////////////////////////////////////////

class AudioMsg &operator>> (AudioMsg &aumsg, GoalShotModuleInfo &info)
{
  aumsg >> info.valid;
  if (info.valid);
  // aumsg >> // further data structures to read ;"
  return aumsg;
}
