//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module goalshotmodule
//
// \$Id: goalshotmodule.h,v 1.8 2001/08/06 02:01:31 satan Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////


#ifndef __GOALSHOTMODULE_H__
#define __GOALSHOTMODULE_H__

#include "grade.h"
#include "goalshoteval.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "mcommand.h"


#include <strstream.h>


class GoalShotModuleInfo;


// the class GoalShotModule
///////////////////////////////

class GoalShotModule
{
public:
  // constructor of module GoalShot 
  GoalShotModule(bool isGoalie, GoalShotEval *_goalshoteval);

  // destructor of module GoalShot 
  ~GoalShotModule();
  
  // feed information about new world and tactics into module
  bool feedbackWorld(const Snapshot* _worldSnap, 
		     TacticalSnapshot* _tacticalSnap);

  // Feed audio information into module. This method is called for
  // each audio message. audioMsg.getSource() is player No or 0 for coach
  // Audiomsg audioInfo is acting like a stream.
  // usage: audioMsg >> auInfo (where GoalShotModuleInfo auInfo holds 
  // the information for module GoalShotModule in other players)"
  bool feedbackAudio(AudioMsg &audioMsg);

  // Do the evaluation for the next action
  // For definition of Grade: see grade.h
  Grade evaluate(AdvSkills* skills,
		 const Snapshot* _worldSnap, TacticalSnapshot* _tacticalSnap);

  /** Decide whether a turn head should be done.
   * This is always called after act, so the data of act and
   * evaluate should be used to calculate the TurnHeadCommand
   * which is then executed by the Mastercontrol.
   * skills should only be used to calculate the command,
   * NOT TO SEND IT!!
   */
  Grade turnHeadEvaluate(AdvSkills* skills, AdvSkills::TurnHeadCommand &);

  // Perform the action for this module  
  bool act(AdvSkills* skills,
	   const Snapshot* _worldSnap, TacticalSnapshot* _tacticalSnap);

  // description of stream AudioMsg: see above. Here you should implement:
  // audioMsg << audioMsg;
  bool getSayMsg(AudioMsg &audioMsg);

private:
  // some constants this class needs
  //////////////////////////////////////

  
  // to store the worldmodel
  const Snapshot *worldSnap;

  // to store tactical infos
  TacticalSnapshot *tacticalSnap;

  // is the player the goalie ?
  bool goalie;       

  // elder things will not be considered
  int maxAge;   

  // some variables this class needs
  //////////////////////////////////////

  // the goalshoteval-instance
  GoalShotEval *goalshoteval;

  // to store the messages that shall be sent
  GoalShotModuleInfo *sendInfo;

  // [C]
  // did the player prepare a superkick and must do a rekick?
  bool mustRekick;   

  // [C]
  // did the player a goalkick, that has been finished?
  bool finishedKick;

  // time of the last snapshot the module received;
  // to check if feedback, evaluate and act have the same judgement-basis
  int lastSnapshotTime;

  // the player himself
  WorldPlayer me;

  // [C]
  // the time the player has to prepare his kick-command
  int availableKickTime;
  
  // which point in the goal should be used to do the goal kick
  Vector2d goalTarget;

  // [C]
  // enough time to use a superkick ?
  bool useSuperKick;

  // [C]
  // enough time to do a fixed kick ?
  bool useFixedKick;

  // for remembering a calculated kick-command
  AdvSkills::ExtKickCommand kickParams;
  

  // some private methods
  /////////////////////////////////////////

  // updates the above datastorage of the module by means of a snapshot
  bool updateDataStorage();
  
  // [C]
  // calculates if a given target is reachable with a kick that only has
  // availableTime to do the kick
  bool targetReachable(AdvSkills *skills, 
		       Vector2d target, int availableTime);


  // determines the best corridor for the player and evaluates it; 
  // returns the grade for the best corridor and the belonging target
  Grade searchAndEvalBestCorridor(AdvSkills *skills, 
				  Vector2d &target, 
				  Vector2d leftTarget,
				  Vector2d middleTarget,
				  Vector2d rightTarget,
				  int availableTime);
  

}; // class GoalShotModule


// Class to keep information for modules of other agents.
//////////////////////////////////////////////////////////////////////

class GoalShotModuleInfo
{
public:

  // constructor of Info class for module GoalShotModule
  GoalShotModuleInfo(bool _valid);
  ~GoalShotModuleInfo();

  // flag to determine if instance of GoalShotModuleInfo is valid
  bool valid;
  
}; // class GoalShotModuleInfo

// put information to AudioMsg-Stream 
class AudioMsg &operator<< (AudioMsg &aumsg, const GoalShotModuleInfo &info);

// fill with information from AudioMsg-Stream 
class AudioMsg &operator>> (AudioMsg &aumsg, GoalShotModuleInfo &info);

#endif











