//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "passmodule.h"
#include "passeval.h"
#include "goalshoteval.h"

// the TRACEs
/////////////////////

#undef USEPASSTRACE
#undef USETESTPASSTRACE
#undef USEDRAWTRACE

// comment out the following line to disable trace
//  #define USEPASSTRACE
//  #define USETESTPASSTRACE
//  #define USEDRAWTRACE

#ifdef USEPASSTRACE
#define PASSTRACE(cmd) cmd
#warning PassModule debug trace PASSTRACE is activated!
#else
#define PASSTRACE(cmd)
#endif


#ifdef USETESTPASSTRACE
#define TESTPASSTRACE(cmd) cmd
#warning PassModule debug trace TESTPASSTRACE is activated!
#else
#define TESTPASSTRACE(cmd)
#endif


#ifdef USEDRAWTRACE
#define DRAWTRACE(cmd) cmd
#warning PassModule debug trace DRAWTRACE is activated
#else
#define DRAWTRACE(cmd)
#endif


/////////////////////////////
//
// class PassModule
//
///////////////////////////

////////////////////////////
//
// Constructor
//
///////////////////////////

PassModule::PassModule(PassEval* _passeval, const GoalShotEval* _goalshoteval):
  callMeAgain(false),             // why should this module be recalled?
  passeval(_passeval),
  goalshoteval(_goalshoteval),
  lastSnapshotTime(-1),           // module has never seen a snaphshot
  time_for_kick(0),
  target_selected(false),
  look_act(false),
  look_for_ball(false),
  // [C] aeh, hm, nun, ja? 
  look_for_ball_time(-3)   // hope that's enough
{
  turnHeadGrade.set_IMPOSSIBLE();
}

////////////////////////////
//
// Destructor
//
///////////////////////////

PassModule::~PassModule()
{}  

//////////////////////////////////////////////////////////////////////
//
// public methods
//
//////////////////////////////////////////////////////////////////////

////////////////////////////
//
// bool PassModule::feedbackWorld
//
///////////////////////////

bool PassModule::feedbackWorld(const Snapshot* _worldSnap, 
			       TacticalSnapshot* _tacticalSnap)
{
  worldSnap = _worldSnap;
  tacticalSnap = _tacticalSnap;

  // just to suppress the unused-parameter-warnings
  if (tacticalSnap);
  
  // if pass is over check if it failed
  if (look_for_ball)
    {
      // if ball velocity does not point to target something must be wrong
      // [C] 0.1?
      if ( fabs( angleBetween( worldSnap->ball.v, 
			       target_pos - worldSnap->my.pos ) ) > PI_4 ||
	   worldSnap->ball.v.getLength() < 0.1 )   // collision?
	{
	  // tactical action can not stay TA_PASS
	  if( tacticalSnap->ourAction.state == TA_PASS &&
	      tacticalSnap->ourAction.performer == worldSnap->my.number )
	    {
	      PASSTRACE({ostrstream ost;
	      ost << " PASS: Pass seems to have gone bananas -> change tacticalAction" 
		  << char(0); 
	      Comm::sendMCmd(MLog(ost.str()));
	      });

	      tacticalSnap->ourAction.state = TA_UNKNOWN;
	      tacticalSnap->ourAction.quality = TQ_GUESSED;
	      tacticalSnap->ourAction.startingTime = worldSnap->time;
	      tacticalSnap->ourAction.performer = UNKNOWN;
	      tacticalSnap->ourAction.receiver = UNKNOWN;
	    }
	}
    }

  return true;
}

///////////////////////////////////
//
// bool PassModule::feedbackAudio
//
///////////////////////////////////

bool PassModule::feedbackAudio(AudioMsg &audioMsg)

{
  bool hasInfo;

  audioMsg >> hasInfo;

  if (hasInfo) {
    cerr << "PASS: received message but never sent one!";
  }

  return true;
}

/////////////////////////////////
//
// Grade PassModule::evaluate
//
////////////////////////////////

Grade PassModule::evaluate(AdvSkills *skills, const Snapshot *_worldSnap,
                           TacticalSnapshot *_tacticalSnap)
{
  // deactivate turnHeadGrade
  ///////////////////////////////
  turnHeadGrade.set_IMPOSSIBLE();

  // deactivate the last target
  ///////////////////////////////
  target_selected = false;

  // just to suppress the unused-parameter-warnings
  if (tacticalSnap);
 

  Grade grade;  // storage for the evaluated grade
  
  // do we need a rekick?
  ///////////////////////////////
  if (callMeAgain && skills->kickSafe(worldSnap)) {
    // action: rekick 
    target_selected = true; // just for safety!
    PASSTRACE({ostrstream ost;
    ost<< "PASS: eval -> rekick needed => ABSOLUTE" 
       << char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });    
    grade.set_ABSOLUTE();
    return grade;
  }
   
  // thought a rekick is needed, but the ball is not kickable
  // kick finished surprisingly or ball is lost!?
  if (callMeAgain && !skills->kickSafe(worldSnap)) {
    callMeAgain = false;
    look_for_ball = false;

    PASSTRACE({ostrstream ost;
    ost	<< "PASS: eval -> don't know if kick is ready or ball is lost "
	<< lastSnapshotTime 
	<< " don't look for ball"              
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });

    look_act = false;
    target_selected = false;
  }

  // time for looking after ball is over
  // [C] 2?
  if ( (look_for_ball) && (look_for_ball_time + 2 < lastSnapshotTime) ) 
    {
      look_for_ball = false;
      PASSTRACE({ostrstream ost;
      ost << "PASS: eval -> don't want to look for the ball anymore"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });

    }

  look_act = false;
  // we have to look for the ball
  // [C] 2?
  if ( (look_for_ball) && (look_for_ball_time + 2 >= lastSnapshotTime) ) 
    {
      PASSTRACE({ostrstream ost;
      ost << "PASS: eval -> want to look for ball => ABSOLUTE"
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
      look_act = true;
      // set the params for turning the head, tell them to the master control
      // in turnHeadEvaluate
      lookAtAngle = worldSnap->my.bodyDir;
      grade.set_ABSOLUTE(); 
      return grade;
    }
    
  // a priori no emergency kick situation
  em_case = false;
 
  // technical impossibility
  ///////////////////////////////
  if ( !skills->kickSafe(worldSnap) ) { 
    
    PASSTRACE({ostrstream ost;
    ost << "PASS: eval -> can't kick the ball => IMPOSSIBLE"
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });  
    
    callMeAgain = false;     // no recall needed
    grade.set_IMPOSSIBLE(); 

    // evtl. ist der Ball im Fuehlradius, etc pp.  ...
    // dann wird ein Pass evaluiert und in die Richtung
    // des Passziels geschaut (turnHead) 
    // [C] Alter des Balls!!!
    if ( (dist(worldSnap->ball.pos, worldSnap->my.pos) < VISIBLE_DISTANCE) &&
	 (worldSnap->ball.state == VISUAL) && (worldSnap->ball.age <= 3)
       ) 
      {

	PASSTRACE({ostrstream ost;
	ost << "PASS: eval -> feeling ball & ball is save; try to look in the right direction"
	    << char(0);
	Comm::sendMCmd(MLog(ost.str()));
	});

	// time available for kick
	time_for_kick = passeval->check_for_safekick(worldSnap, 
					       worldSnap->my.number);
 
	// as we probably want to pass in what follows, we  
	// gather some information first:
  
	// fill the propertyof-array and the rangepassof-array
	getPasses( (worldSnap->my.isGoalie == GOALIE_TRUE) ? true : false);
  
	PropOf prop;
	prop.grade.set_BAD();
	// try to get best p-pass and best r-pass as partner
	for (int l = 1; l <= 11; l++) {
	  if (pprops[l].perhaps)
	    if (passeval->betterPass(worldSnap, pprops[l], prop)) {
	      prop = pprops[l];
	      target_breed = true;
	    }
	  if (rprops[l].perhaps)
	    if (passeval->betterPass(worldSnap, rprops[l], prop)) {
	      prop = rprops[l];
	      target_breed = false;
	    }
	} // for  
	
	if ( prop.grade.is_better_than(Grade::POOR) )
	  {
	    DRAWTRACE(Comm::sendMCmd(MDLine(worldSnap->my.pos,prop.pos)));
	    
	    Angle bodyAngleToTarget = worldSnap->my.bodyAngleTo( WorldObject(prop.pos) );
	    // [C] falls MINNECKMOMENT, MAXNECKMOMENT mal nicht mehr -180, +180
	    // sind, bitte aendern !!!
	    if ( ( bodyAngleToTarget < MAXNECKANG ) && 
		 ( bodyAngleToTarget > MINNECKANG ) ) 
	      {
		turnHeadGrade.set_PERFECT();
		lookAtPos = prop.pos;
	      }
	  }

      } // ball im Fuehlradius
  
    return grade;
  } // ball not kickable

  // just here we're able to KICK THE BALL
  ///////////////////////////////////////////
 
  // information gathering
  /////////////////////////////

  // time available for kick
  time_for_kick = passeval->check_for_safekick(worldSnap, 
					       worldSnap->my.number);
 
  // as we probably want to pass in what follows, we  
  // gather some information first:
  
  // fill the propertyof-array and the rangepassof-array
  getPasses( (worldSnap->my.isGoalie == GOALIE_TRUE) ? true : false );
  
  PropOf prop;
  prop.grade.set_BAD();

  // try to get best p-pass and best r-pass as partner
  for (int l = 1; l <= 11; l++) {
    if (pprops[l].perhaps)
      if (passeval->betterPass(worldSnap, pprops[l], prop)) {
	prop = pprops[l];
	target_breed = true;
      }
    if (rprops[l].perhaps)
      if (passeval->betterPass(worldSnap, rprops[l], prop)) {
	prop = rprops[l];
	target_breed = false;
      }
  }
  
  // no p- and no r-pass found, so it's an emergency-kick-away-do-something
  if (prop.grade.is_BAD()) {
    em_case = true;
    em_vec = get_ePass(worldSnap->playMode);

    PASSTRACE({ostrstream ost;
    ost << " PASS: eval -> no pass found, emergency-kick-away considered => BAD" 
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });

    grade.set_BAD();
    return grade;
  }
  
  lockPass(prop);  

  // downgrade pass if player is the goalie and should catch instead
  if( prop.grade.is_PERFECT()
      && worldSnap->my.number == worldSnap->ourGoalieNo
      && worldSnap->my.inOurPenaltyArea()
      && worldSnap->playMode == PLAY_MODE_PLAY_ON )
    {
      prop.grade.downgrade();
    }

  // if there might have been a ball collision let ballhandling module do the job
  if (worldSnap->collState != NONE)
    if (prop.grade.is_PERFECT())
      prop.grade.set_GOOD();


  Angle wantedToTurnAngle =  worldSnap->my.faceAngleTo( WorldObject(prop.pos) );
  Angle viewAngle = worldSnap->senseBody.getViewAngle();

  // wir haben einen mindestens mit gut bewerteten Pass; gucken wir fr diesen
  // schon gut genug auf das target ?
  if( (( wantedToTurnAngle < -viewAngle ) ||
       ( wantedToTurnAngle > viewAngle  ))
      && prop.grade.is_better_than(Grade::AVERAGE) 
      && (time_for_kick >= 2) )
    {
      // target liegt nicht im inneren Sichtbereich
      // der ausgesuchte Pass ist GOOD oder besser
      // der Spieler hat mindestens 2 Zeitschritte Zeit

      Angle bodyAngleToTarget = worldSnap->my.bodyAngleTo( WorldObject(prop.pos) );
      // [C] falls MINNECKMOMENT, MAXNECKMOMENT mal nicht mehr -180, +180
      // sind, bitte aendern !!!
      if ( ( bodyAngleToTarget < MAXNECKANG ) && 
	   ( bodyAngleToTarget > MINNECKANG ) ) 
	{
	  // drehen ist moeglich
	  lookAtPos = prop.pos;
	  turnHeadGrade.set_PERFECT();
	  // downgrade pass - but not near opponent goal
	  // in this case risk something 
	  if( worldSnap->my.pos.getX() < FIELD_LENGTH / 4.0 )
	    grade.set_AVERAGE();
	  else
	    grade = prop.grade;
	  	  
	  PASSTRACE({ostrstream ost;
	  ost << "PASS: eval -> target not in view cone; turn Head; grade set to "
	      << grade << char(0);
	  Comm::sendMCmd(MLog(ost.str()));
	  });
	  
	  DRAWTRACE(Comm::sendMCmd(MDPosition(prop.pos)));
	  
	  return grade;
	    
	}// turning the head in the right direction is possible

      else
	{
	  // drehen des Kopfes reicht nicht, Drehung des Koerpers ist gefragt
	  // => Ballhandling dranlassen
	  grade.set_BAD();
	  PASSTRACE({ostrstream ost;
	  ost << "PASS: eval -> target not in view cone; turn Head not possible; grade = "
	      << grade << char(0);
	  Comm::sendMCmd(MLog(ost.str()));
	  });
	  DRAWTRACE(Comm::sendMCmd(MDPosition(prop.pos)));
	  return grade;
	} // turning of head is not possible
    } // target position is not in view cone
  else
    {
      PASSTRACE({ostrstream ost;
      ost << " PASS: eval -> found a partner:" 
	  << prop
	  << " time for kick : " 
	  << time_for_kick
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      if (!target_breed)
	{
	  ost << "it is a rangepass!" << char(0) ;
	  Comm::sendMCmd(MLog(ost.str()));
	}
      });
      
      grade = prop.grade;  
      DRAWTRACE(Comm::sendMCmd(MDPosition(prop.pos)));
      
      return grade;
    } // target position in view cone

}  // PassModule::eval


////////////////////////////////
//
// PassModule::turnHeadEvaluate
//
////////////////////////////////

Grade PassModule::turnHeadEvaluate(AdvSkills *skills, 
				   AdvSkills::TurnHeadCommand &thc)
{
  Grade g;

  if(look_act)
    {
      g.set_PERFECT();
      thc = skills->turnHead(worldSnap, lookAtAngle);
      PASSTRACE({ostrstream ost;
      ost << "PASS: turnHeadEvaluate -> want to turn head to body-direction"
	  << " grade = " << g
	  << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }
  else
    {
      if ((tacticalSnap->ourAction.state!=TA_REACT) ||
	  (turnHeadGrade.is_worse_than(Grade::AVERAGE)))
	g = turnHeadGrade;
      else g.set_AVERAGE();

      thc = skills->turnHead(worldSnap, WorldObject(lookAtPos));
    }

  return g;
}


////////////////////////////
//
// PassModule::act
//
///////////////////////////

bool PassModule::act(AdvSkills *skills, const Snapshot *_worldSnap,
                     TacticalSnapshot *_tacticalSnap)
{

  // we have to do an emergency-kick-away
  if (em_case) {
    PASSTRACE({ostrstream ost;
    ost << "PASS: act -> performed emergency-kick-away; direction = "
	<< RAD_TO_DEG( em_vec.getAngle() ) 
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });
    kickparams = skills->kick(worldSnap, em_vec, AdvSkills::KICKAWAY);
    skills->exec(worldSnap, kickparams);
    return true;
  }

  if (look_act) {  // look for the ball; don't do anything else
    
    
    // set the params for turning
    AdvSkills::TurnCommand turnparams 
      = skills->turn(worldSnap, 
		     (worldSnap->ball.pos - worldSnap->my.pos).getAngle());
    skills->exec(worldSnap, turnparams);

    PASSTRACE({ostrstream ost;
    ost << "PASS: act -> turned body to ball to look at it"
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });

    DRAWTRACE(Comm::sendMCmd(MDLine(target_source,target_pos)));
    DRAWTRACE(Comm::sendMCmd(MDVLine(worldSnap->ourOffsideLine(MAXAGE))));
    DRAWTRACE(Comm::sendMCmd(MDCorridor(target_source,target_pos,target_corridor,PI_6)));

    return true;
  }

  if(!(skills->kickSafe(worldSnap))){
    callMeAgain = false;
    skills->noAction();
    PASSTRACE({ostrstream ost;
    ost << "PASS: act -> did no action because ball not kickable"
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });
    return false;
  }

  // the kick is not finished, so perform a rekick
  if (callMeAgain) {
    kickparams = skills->rekick(worldSnap, kickparams);
    skills->exec(worldSnap, kickparams);
    if (kickparams.state == AdvSkills::KFINISHING) {  // the kick is finished
      callMeAgain = false;
      look_for_ball = true;                  // look after the ball
      look_for_ball_time = lastSnapshotTime; // from now on (for 2 cycles)
    }

    DRAWTRACE(Comm::sendMCmd(MDLine(target_source,target_pos)));
    DRAWTRACE(Comm::sendMCmd(MDVLine(worldSnap->ourOffsideLine(MAXAGE))));
    DRAWTRACE(Comm::sendMCmd(MDCorridor(target_source,target_pos,target_corridor,PI_6)));

    PASSTRACE({ostrstream ost;
    ost << "PASS: act -> did a rekick" 
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    if (kickparams.state == AdvSkills::KFINISHING)
      ost << "finished" << char(0);
    else 
      ost << " not finished" << char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });
    
    return true;             
  }

  if (target_selected) {
    target_source = worldSnap->my.pos;

    double end_vel;   // ball velocity when it reaches the partner

    // get the kickparams
    if (target_breed)
      // playerpass => ball should have vel. == 0.8 when reaching partner
      // [C] MAXPOWER usw.
      end_vel = ( MAXPOWER * KICK_POWER_RATE /
		  ( dist( target_source, target_pos ) 
		    < passeval->SHORTPASS_LENGTH ? 4 : 2 ) );
    else
      // rangepass => ball should leave with calculated velocity
      // [C] MAXPOWER usw.
      end_vel = min( target_vel, MAXPOWER * KICK_POWER_RATE /
		     ( dist( target_source, target_pos ) 
		       < passeval->SHORTPASS_LENGTH ? 4 : 2 ) );

    kickparams = skills->kick(worldSnap, end_vel, target_pos, time_for_kick);
    
    // perform the kick
    skills->exec(worldSnap, kickparams);

    // report the kick
    tacticalSnap->reportOurTacticalAction( TA_PASS, worldSnap->my.number, 
					   target_number );

    // if we want to look for the ball later, here's the pos
    look_for_ball_pos = target_pos;

    if (kickparams.state != AdvSkills::KFINISHING) {           
      // pass is not finished
      callMeAgain = true;                                      
      // prepare for rekick
      PASSTRACE({ostrstream ost;
      ost << "PASS: not finished" << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }
    else {                                                     
      // kick is ready by now 
      look_for_ball = true;                                    
      // prepare for looking after ball
      look_for_ball_time = lastSnapshotTime;
      PASSTRACE({ostrstream ost;
      ost << "PASS: finished" << char(0);
      Comm::sendMCmd(MLog(ost.str()));
      });
    }
    
    DRAWTRACE(Comm::sendMCmd(MDLine(target_source,target_pos)));
    //DRAWTRACE(Comm::sendMCmd(MDCircle(target_pos, passeval->DANGERZONE)));
    DRAWTRACE(Comm::sendMCmd(MDVLine(worldSnap->ourOffsideLine(MAXAGE))));
    DRAWTRACE(Comm::sendMCmd(MDCorridor(target_source,target_pos,target_corridor,PI_6)));
    
    PASSTRACE({ostrstream ost;
    ost << " PASS: act -> trying to pass to " << target_number 
	<< " at pos" << target_pos 
	<< " with " << time_for_kick << " cycles time to kick"
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    if (!target_breed)
      {
	ost << "this is a rangepass" 
	    << char(0);
	Comm::sendMCmd(MLog(ost.str()));
      }
    });

    return true;
  }

  PASSTRACE({ostrstream ost;
  ost << "PASS: act -> shit, no action sent" << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  }); 
  skills->noAction(); // to be changed
  return false;
} // PassModule::act

////////////////////////////
//
// PassModule::getSayMsg
//
///////////////////////////

bool PassModule::getSayMsg(AudioMsg &audioMsg)
{
  bool sendSomething = false;

  if (sendSomething) {

    audioMsg.markWrite();
    
    audioMsg << true;

    // audioMsg << sendInfo;
    
    // check if error occurred
    if( !audioMsg.getState() )
      {
        cerr << "PASS: Stream overflow: " << audioMsg ;
        audioMsg.rollbackWrite().clearState();
	audioMsg << false;
      }
  } 
  else {
    audioMsg << false;
  }

  return true;;
}

/////////////////////////////////////////////////////////////////////////
//
// private methods
//
/////////////////////////////////////////////////////////////////////////


////////////////////////////////////
//
// bool PassModule::getPasses
//
///////////////////////////////////

bool PassModule::getPasses(bool isGoalie)
{
  for (int i = 1; i <= 11; i++) {
    TESTPASSTRACE({ostrstream ost;
    ost << " These are the passes for player " << i 
	<< char(0);
    Comm::sendMCmd(MLog(ost.str()));
    });

    pprops[i].perhaps = false;
    pprops[i].grade.set_BAD();
    rprops[i].perhaps = false;
    rprops[i].grade.set_BAD();
    
    // we don't want to pass to ourselves
    if( i == worldSnap->my.number )
      {
	TESTPASSTRACE({ostrstream ost;
	ost << " This is my own number!" << char(0) ;
	Comm::sendMCmd(MLog(ost.str()));
	});
      }
    else
      // we don't want to pass to players that are older than MAXAGE
      if( worldSnap->ourTeam(i).age > MAXAGE )
	{
	  TESTPASSTRACE({ostrstream ost;
	  ost << " This player is too old!" << char(0);
	  Comm::sendMCmd(MLog(ost.str()));
	  });
	}
      else
	// we don't want to pass to our goalie
	if( i == worldSnap->ourGoalieNo )
	  {
	    TESTPASSTRACE({ostrstream ost;
	    ost << " This player is the goalie!" << char(0);
	    Comm::sendMCmd(MLog(ost.str()));
	    });
	  }
	else
	  // if we do not know our goalie yet we don't want to pass to players
	  // that are close to our goal ([C] FIELD_LENGTH/6.0)
	  if( worldSnap->ourGoalieNo == UNKNOWN && 
	      worldSnap->ourTeam(i).pos.getX() < - FIELD_LENGTH/6.0 )
	    {
	      TESTPASSTRACE({ostrstream ost;
	      ost <<  " I don't know our goalie number"
		  << "and this player is too close to our goal!"
		  << char(0);
	      Comm::sendMCmd(MLog(ost.str()));
	      });
	    }
	  else
	    {
	      // in all other cases: get player passes and range passes
	      passeval->get_pPass(true, worldSnap, worldSnap->my.number, 
				  worldSnap->ourTeam(i).pos,
				  i, pprops[i], isGoalie);
	      passeval->get_rPass(true, worldSnap, worldSnap->my.number, i, 
				  rprops[i], isGoalie);
	      TESTPASSTRACE({ostrstream ost;
	      ost << " Evaluation of playerpass:" 
		  << pprops[i]
		  << " Best rangepass:" 
		  << rprops[i]
		  << char(0);
	      Comm::sendMCmd(MLog(ost.str()));
	      });
	    }
  }

  return true;
} // getPasses

////////////////////////////
//
// PassModule::get_ePass
//
///////////////////////////

Vector2d PassModule::get_ePass(play_mode playMode)
{
  // [C] der ganze e-Pass-Kram ist wohl Konstanten-orientiert programmiert! 
  // evtl. aendern?
  // 9 possible directions for emergency kick
  Angle kickangle[9] = { -PI_2, -PI_4 - PI_8, -PI_4, -PI_8, 0.0, PI_8, 
			 PI_4, PI_4 + PI_8, PI_2 };
  if (playMode == PLAY_MODE_CORNER_KICK_US)
    {
      kickangle[0] = -PI;
      kickangle[1] = -PI + PI_8;
      kickangle[2] = -PI + PI_4;
      kickangle[3] = -PI_2 - PI_8;
      kickangle[4] = -PI_2; 
      kickangle[5] =  PI_2;
      kickangle[6] =  PI_2 + PI_8;
      kickangle[7] =  PI_2 + PI_4; 
      kickangle[8] =  PI - PI_8;
    }
  

  // evaluations of the 9 possible directions
  Grade dir[9];
  for( int i = 0; i < 9; i++ )
    dir[i].set_PERFECT();

  Vector2d my_pos = worldSnap->my.pos;

  const int AREA_LEFT   = 1;
  const int AREA_MIDDLE = 2;
  const int AREA_RIGHT  = 3;

  const int SI_DEFENSIVE = 1;
  const int SI_MIDFIELD  = 2;
  const int SI_OFFENSIVE = 3;

  int my_area = ( ( my_pos.getY() < -FIELD_WIDTH / 6.0 ) ? AREA_LEFT :
		  ( my_pos.getY() > FIELD_WIDTH / 6.0 ) ? AREA_RIGHT : 
		  AREA_MIDDLE );
  
  int my_situation = ( ( my_pos.getX() < -FIELD_LENGTH / 6.0 ) ? SI_DEFENSIVE :
		       ( my_pos.getX() > FIELD_LENGTH / 6.0 ) ? SI_OFFENSIVE : SI_MIDFIELD );
  
  // suppress certain directions due to play mode, situation and my position
  if(playMode == PLAY_MODE_FREE_KICK_US)
    {
      switch( my_situation )
	{
	case SI_DEFENSIVE :
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	      dir[3].set_BAD(); dir[4].set_BAD(); dir[7].set_BAD();
	      dir[8].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      break;
	    case AREA_RIGHT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[4].set_BAD();
	      dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD();
	      dir[8].set_BAD();
	      break;
	    }
	  break;
	case SI_MIDFIELD:
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	      dir[3].set_BAD(); dir[4].set_BAD(); dir[8].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      break;
	    case AREA_RIGHT:
	      dir[0].set_BAD(); dir[4].set_BAD(); dir[5].set_BAD(); 
	      dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    }
	  break;
	case SI_OFFENSIVE:
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	      dir[3].set_BAD(); dir[4].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      break;
	    case AREA_RIGHT:
	      dir[4].set_BAD(); dir[5].set_BAD(); 
	      dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    }
	  break;
	}
    }
  else
    if( playMode == PLAY_MODE_CORNER_KICK_US)
      {
	if (AREA_LEFT)
	  {
	    dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	    dir[4].set_BAD(); dir[3].set_BAD();
	  }
	else
	  // right corner
	  {
	    dir[4].set_BAD(); dir[5].set_BAD(); dir[6].set_BAD();
	    dir[7].set_BAD(); dir[8].set_BAD();
	  }
      }
    else
      // player stands in the field
      {
	switch( my_situation )
	  {
	  case SI_DEFENSIVE :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); 
		dir[8].set_BAD();
		break;
	      case AREA_MIDDLE :
		break;
	      case AREA_RIGHT  :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); 
		dir[3].set_BAD();
		break;
	      }
	    break;
	    
	  case SI_MIDFIELD :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
		dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_MIDDLE :
		dir[0].set_BAD(); dir[1].set_BAD();
		dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_RIGHT  :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
		dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      }
	    break;
	    
	  case SI_OFFENSIVE :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); 
		dir[3].set_BAD();
		break;
	      case AREA_MIDDLE :
		dir[0].set_BAD(); dir[1].set_BAD();
		dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_RIGHT  :
		dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); 
		dir[8].set_BAD();
		break;
	      }
	    break;
	    
	  default :   // should not happen
	    break;
	  }
      }

  // first check if a short and narrow corridor is free
  // [C] lauter Konstanten fuer die Korridorbreiten und -winkel!
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( worldSnap->otherTeamInCorridor( my_pos, my_pos 
					  + radAng2Vector2d( 15.0, kickangle[i] ), 
					  5.0, PI_2 ) )
	dir[i].set_POOR();
  
  // ...then try it with a longer and broader one
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( worldSnap->otherTeamInCorridor( my_pos, my_pos 
					  + radAng2Vector2d( 17.0, kickangle[i] ), 
					  8.0, PI_2 ) )
	dir[i].set_AVERAGE();
  
  // ...finally test the longest and broadest corridor
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( worldSnap->otherTeamInCorridor( my_pos, my_pos 
					  + radAng2Vector2d( 20.0, kickangle[i] ), 
					  10.0, PI_2 ) )
	dir[i].set_GOOD();
  
  // search for best grade
  int bestDir = ( my_pos.getY() < 0 ? 0 : 8 );
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_better_than( dir[bestDir] ) )
      bestDir = i;
  
  DRAWTRACE(Comm::sendMCmd(MDLine(my_pos,my_pos 
				  + radAng2Vector2d(15.0,kickangle[bestDir]))));
      
  PASSTRACE({ostrstream ost;
  ost << " PASS: get_ePass -> emergency kick into direction " 
      << kickangle[bestDir]
      << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  });

  // determine kick vector
  return radAng2Vector2d( BALL_SPEED_MAX, kickangle[bestDir] );
}

////////////////////////////
//
// PassModule::lockPass
//
///////////////////////////

Grade PassModule::lockPass(PropOf prop)
{
  target_selected = true;
  target_number = prop.number;
  target_pos = prop.pos;
  target_vel = prop.vel;
  target_corridor = prop.corridor;

  return prop.grade;
}

