//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// \$Id:
// \$manager: gauer$
//
//////////////////////////////////////////////////////////////////////


#include "standardsituationmodule.h"
#include "Config.h"


int
StandardSituationModule::
calcPlayersAndPos4FreeKickThem(const Snapshot *worldSnap, 
			       const TacticalSnapshot *ts,
			       int players[],
			       Vector2d target[]) const
{

  /* ??? keep wall building? */
  int count(0);
  // if the ball is close to our goal
  // [C]*
  if (sqr(12.) < sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) &&
      sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) < sqr(30.))
    {
      // calc the postitions the players are going to
      
      // The left and the right edges of our goal
      Vector2d leftPost(worldSnap->ourGoal.pos.getX(),
			worldSnap->ourGoal.pos.getY()+GOAL_WIDTH/2.0),
	rightPost(worldSnap->ourGoal.pos.getX(),
		  worldSnap->ourGoal.pos.getY()-GOAL_WIDTH/2.0);
      
      // calc the positions they will go to
      // on the lines from the ball to the left (right) edge of the goal
      target[0] = (leftPost - worldSnap->ball.pos); 
      target[1] = (rightPost - worldSnap->ball.pos);
      double angleBetweenPosts = angleBetween(target[0], target[1]);
      // [C]
      double range = 1./3.;  // How far from the circle are we
      // a+range(b-a) == ...
      target[0].stretch(FREE_KICK_FORBIDDEN_AREA/norm(target[0]));
      target[0] += worldSnap->ball.pos;
      target[0].stretch(1-range);
      target[0] += range*leftPost;
      target[1].stretch(FREE_KICK_FORBIDDEN_AREA/norm(target[1]));
      target[1] += worldSnap->ball.pos;
      target[1].stretch(1-range);
      target[1] += range*rightPost;
  
      // if there is enough space between the two positions
      // move players a little together
      // [C]*
      if (angleBetweenPosts > DEG_TO_RAD(18.))
	{
	  target[0] -= worldSnap->ball.pos;
	  target[0].turn(DEG_TO_RAD(-4.));
	  target[0] += worldSnap->ball.pos;
	  target[1] -= worldSnap->ball.pos;
	  target[1].turn(DEG_TO_RAD(4.));
	  target[1] += worldSnap->ball.pos;
	}
      
      DRAWTRACE( POSITION(target[0]) );
      DRAWTRACE( POSITION(target[1]) );
      
      count = calcClosestDesiredPlayersTo(target, 2, worldSnap, ts, players);

    }
  else 
    {
      if (sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) > sqr(minWallDist))
	{
	  // the ball is not close to our goal
	  DECISIONTRACE( " ball is not close to our Goal" );

	  int p = calcOurKicksThemPositions(worldSnap, ts, target);

	  DECISIONTRACE(" found " << p << " dangerous passes");

	  count = calcClosestDesiredPlayersTo(target, p, worldSnap, ts, players);
	  
	}
      else
	{
	  DECISIONTRACE( " Ball is closer than " << minWallDist << " to our goal! " );
	  count = 0;
	}
    }
  return count;
}


int
StandardSituationModule::
calcPlayersAndPos4CornerKickThem(const Snapshot *worldSnap, 
				 const TacticalSnapshot *ts,
				 int players[],
				 Vector2d target[]) const
{
  /* ??? check wich of the other players can make a goal 
     and block a pass to him */

  int p = calcOurKicksThemPositions(worldSnap, ts, target);
  
  DECISIONTRACE(" found " << p << " dangerous passes");
  
  return calcClosestDesiredPlayersTo(target, p, worldSnap, ts, players);
}

int
StandardSituationModule::
calcPlayersAndPos4KickInThem(const Snapshot *worldSnap, 
			     const TacticalSnapshot *ts,
			     int players[],
			     Vector2d target[]) const
{
  int p = calcOurKicksThemPositions(worldSnap, ts, target);
  
  DECISIONTRACE(" found " << p << " dangerous passes");
  
  return calcClosestDesiredPlayersTo(target, p, worldSnap, ts, players);
}

bool 
StandardSituationModule::
doingGoalieFreeKickThem(AdvSkills *skills,
			const Snapshot *worldSnap)
{
  DECISIONTRACE( "Balldist: " << worldSnap->my.distTo(worldSnap->ball) );

  // calc goalie position
  Vector2d intersect = calcPossibleGoalPoint(worldSnap);
  Vector2d pos = intersect - worldSnap->ball.pos;
  // [C]
  pos = (1. - 3./norm(pos))*pos + worldSnap->ball.pos;
  
  // if this pos is inside of FREE_KICK_FORBIDDEN_AREA
  if (sqrnorm(pos-worldSnap->ball.pos) <= sqr(FREE_KICK_FORBIDDEN_AREA))
    {
      // translate it PLAYER_SIZE meters in front of it
      pos =
	(pos-worldSnap->ball.pos).stretch((FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE)/
					  norm(pos-worldSnap->ball.pos))+
	worldSnap->ball.pos;
    }
  
  // if this is behind the goal
  if (pos.getX() < -FIELD_LENGTH/2)
    {
      ERRORTRACE( " This freekick is closer than "\
		  "FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE to the goal! :-( " );
      DECISIONTRACE( " Goalie turns to ball " );
      // goalie should do anything (walk to good position near the circle around 
      // the ball or so) not only look to ball
      Snapshot mySnap = *worldSnap, *ws = &mySnap;
      AdvSkills::TurnCommand tcmd =
	skills->turn(worldSnap, 				 
		     worldSnap->my.bodyAngleTo(worldSnap->ball)+
		     worldSnap->my.bodyDir);
      if (tcmd.state != AdvSkills::RFINISHED)
	skills->exec(worldSnap, tcmd);
      else
	skills->noAction();
      ws->my.bodyDir = tcmd.bodyDir;
      thcmd = skills->turnHead(ws, ws->my.bodyAngleTo(ws->ball)+
			           ws->my.bodyDir);
      wantToTurnHead=true;
      return true;
    }

  double stamina = calcUsableStamina(goalieStaminaFactor, worldSnap);
  
  if (!goingBackwardsTo(pos, stamina, skills, worldSnap))
    {
      skills->noAction();
      lookAround(skills, worldSnap, worldSnap->ball.pos);
      isSpecialSituation = true;
      DECISIONTRACE( "Using SPECIAL SITUATION channel" );
    }
  return true;
}

bool 
StandardSituationModule::
doingFieldPlayerKicksThem(AdvSkills *skills,
			  const Snapshot *worldSnap,
			  const int players[],
			  const Vector2d target[], 
			  const int count)

{
  Vector2d targetPos;

  for (int i = 0; i < count; i++)
    if (worldSnap->my.number == players[i] )
      {
	DECISIONTRACE( " I am player " << i+1 << " out of " << count << " players");
	targetPos = target[i];
      }

  Vector2d oponent, T = targetPos + 10*(targetPos-worldSnap->ball.pos);
  double d(sqr(NOWHERE_DIST));
  for (int i = 1; i < 12; i++) {
    if (worldSnap->otherTeam(i).inCorridor(targetPos,T,2*PLAYER_SIZE) &&
	sqrnorm(worldSnap->otherTeam(i).pos-targetPos) < d) {
      d = sqrnorm(worldSnap->otherTeam(i).pos-targetPos);
      oponent = worldSnap->otherTeam(i).pos;
    }
  }

  DRAWTRACE( POSITION(oponent) );

  // [C]
  if (sqrnorm(oponent-targetPos) > sqr(10) &&
      sqrnorm(targetPos-worldSnap->my.pos) > sqrnorm(oponent-targetPos))
    targetPos = targetPos + .4*(oponent-targetPos);

  DECISIONTRACE( " My targetPos: " << targetPos );

  double stamina(calcUsableStamina(passiveStaminaFactor, worldSnap));
  
  if (sqr(minWallDist) < sqrnorm(worldSnap->ourGoal.pos-worldSnap->ball.pos) &&
      sqrnorm(worldSnap->ourGoal.pos-worldSnap->ball.pos)< sqr(maxWallDist)) {
    DECISIONTRACE( " Ball is dangerous!, Balldist: " 
		   << worldSnap->my.distTo(worldSnap->ball) );
    stamina = calcUsableStamina(wallStaminaFactor, worldSnap);
  }

  if (!goingTo(targetPos, stamina, skills, worldSnap, &(worldSnap->ball.pos),
	       FREE_KICK_FORBIDDEN_AREA+1.5*PLAYER_SIZE))
    {
      if (worldSnap->time%6 == 0) { //[C] ???
	DECISIONTRACE( "reached my pos. want to turn around to update WM" );
	AdvSkills::TurnCommand tcmd = 
	  skills->turn(worldSnap, worldSnap->my.bodyDir+PI);
	if (tcmd.state == AdvSkills::RANGLE ||
	    tcmd.state == AdvSkills::ROK)
	{
	  DECISIONTRACE( " turning around... " );
	  skills->exec(worldSnap, tcmd);
	  thcmd = skills->turnHead(worldSnap, tcmd.bodyDir);
	  wantToTurnHead = true;
	}
	else {  // should never happern
	  DECISIONTRACE( " I allready turned around??? " );
	  skills->noAction();
	  lookAround(skills, worldSnap, worldSnap->ball.pos);
	}
      }
      else {
	DECISIONTRACE( "reached my pos. looking around ball" );
	skills->noAction();
	lookAround(skills, worldSnap, worldSnap->ball.pos);
      }
    }
  return true;
}
