//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// \$Id: stdKicksUs.C,v 1.20 2001/08/01 22:07:27 satan Exp $
// \$manager: gauer$
//
//////////////////////////////////////////////////////////////////////


#include "standardsituationmodule.h"
#include "Config.h"


int 
StandardSituationModule::
calcPos4CornerKickUs(const Snapshot *worldSnap, 
		     Vector2d target[]) const
{
  // the first position is directly behind the ball on the line
  // form ball to (0,0)
  target[0] =  -worldSnap->ball.pos;
  // [C]
  target[0] *= (BALL_SIZE+worldSnap->my.size)/norm(target[0]);
  target[0] = worldSnap->ball.pos - target[0];
  
  return 1;
}

int 
StandardSituationModule::
calcPlayersAndPos4CornerKickUs(const Snapshot *worldSnap, 
			       const TacticalSnapshot *ts,
			       int players[],
			       Vector2d target[]) const
{
  int count = calcPos4CornerKickUs(worldSnap, target);
  count = calcClosestDesiredPlayersTo(target, count, worldSnap, ts, players);

  DRAWTRACE( POSITION(target[0]) );

  return count;
}

int 
StandardSituationModule::
calcPlayersAndPos4KickInUs(const Snapshot *worldSnap, 
			   const TacticalSnapshot *ts,
			   int players[],
			   Vector2d target[]) const
{
  double r(BALL_SIZE+worldSnap->my.size);
  target[0] = worldSnap->ball.pos;
  if (target[0].getY() > 0)
    target[0] += Vector2d(0, r);
  else if (target[0].getY() < 0)
    target[0] -= Vector2d(0, r);
  DRAWTRACE( POSITION(target[0]) );

  return calcClosestDesiredPlayersTo(target, 1, worldSnap, ts, players);
}

int 
StandardSituationModule::
calcPlayersAndPos4FreeKickUs(const Snapshot *worldSnap, 
			     const TacticalSnapshot *ts,
			     int players[],
			     Vector2d target[]) const
{
  double r(BALL_SIZE+worldSnap->my.size);
  target[0] = worldSnap->ball.pos;
  if (target[0].getY() > FIELD_WIDTH/6.)
    target[0] += Vector2d(0, r);
  else if (target[0].getY() < -FIELD_WIDTH/6.)
    target[0] -= Vector2d(0, r);
  else
    target[0] -= Vector2d(r, 0);
  DRAWTRACE( POSITION(target[0]) );

  return calcClosestDesiredPlayersTo(target, 1, worldSnap, ts, players);
}

bool 
StandardSituationModule::
doingFreeKickUs(AdvSkills *skills,
		const Snapshot *worldSnap,
		const int players[],
		const Vector2d target[],
		const int count)
{
  Vector2d targetPos;
  bool iWillKick(false);
  double r(BALL_SIZE), mySize(worldSnap->my.size);
  Vector2d myPos(worldSnap->my.pos), ball(worldSnap->ball.pos);
  
  for (int i = 0; i < count; i++)
    if (worldSnap->my.number == players[i] )
      {
	DECISIONTRACE( " I am player " << i+1 << " out of " << count << " players");
	targetPos = target[i];
	if (i == 0) iWillKick = true;
      }

  DECISIONTRACE( " My targetPos: " << targetPos );
  OTHERTRACE( " My pos: " << myPos );

  if (iWillKick) { //[C]
    r = mySize + worldSnap->my.kickableMargin;
    if (targetPos.getY() > ball.getY())
      targetPos += Vector2d(0, r+PLAYER_SIZE);
    else if (targetPos.getY() < ball.getY())
      targetPos -= Vector2d(0, r+PLAYER_SIZE);
    else
      targetPos -= Vector2d(r+PLAYER_SIZE, 0);

    if (worldSnap->my.inCorridor(ball, targetPos, 4*mySize, PI_2)) {
      DRAWTRACE( CORRIDOR(ball, targetPos, 4*mySize, PI_2) );
      targetPos = target[0];
      r = BALL_SIZE;
    }
    DRAWTRACE( POSITION(targetPos) );
  }

  double stamina(calcUsableStamina(activeStaminaFactor, worldSnap));
  
  if (!goingTo(targetPos, stamina, skills, 
	       worldSnap, &ball, r))
    {
      if (iWillKick) {
	bool wait(true);
	int myNum(worldSnap->my.number);
	Grade passGrade;
	
	for (int pnum = 1; pnum < 12; pnum++)
	  if (pnum != myNum) {
	    passGrade = passEval->ourEval(worldSnap, myNum, pnum);
	    if (passGrade.is_better_than(Grade::AVERAGE) &&
		worldSnap->ourTeam(pnum).age <= 2) 
	      wait = false;

	    DRAWTRACE( LINE(ball, worldSnap->ourTeam(pnum).pos) );
#ifdef USEDRAWTRACE
	    {
	      ostrstream text;
	      text << passGrade << (char)0;
	      DRAWTRACE( TEXT(text.str(), worldSnap->ourTeam(pnum).pos) );
	    }
#endif
	  }
	  else
	    continue;

#ifdef USEDRAWTRACE
  {
    ostrstream text;
    text << "GoalShotEval Grade: " << goalShotEval->ourEval(worldSnap, myPos) << (char)0;
    DRAWTRACE( TEXT(text.str(), Vector2d(-FIELD_LENGTH/2,-FIELD_WIDTH/2)) );
  }
#endif

  OTHERTRACE("next time of EMRG channel: " << worldSnap->nextTimeOfChannel(AUMUL_EMRG));

	if (((waitingTime-- <= 0 || !wait) // make shure we can scream next cycle
	     && worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 == 0) 
	    || goalShotEval->ourEval(worldSnap, myPos).is_better_than(Grade::AVERAGE)
	    || timeLeft <= 0) {
	  stdSitFinished = true;
	  AdvSkills::TurnCommand tcmd = 
	    skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
			 worldSnap->my.bodyDir);
	  if (tcmd.state != AdvSkills::RFINISHED)
	    skills->exec(worldSnap, tcmd);
	  else
	    skills->noAction();
	  thcmd = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
	  wantToTurnHead = true;
	  DECISIONTRACE("Standardsituation finished");
	}
	else {
	  if (wait && worldSnap->time%6 == 0 && fabs(myPos.getY()) < FIELD_WIDTH/6.) { //[C] ???
	    DECISIONTRACE( "reached my pos. want to turn around to update WM" );
	    AdvSkills::TurnCommand tcmd = 
	      skills->turn(worldSnap, worldSnap->my.bodyDir+PI);
	    if (tcmd.state != AdvSkills::RFINISHED) {
	      DECISIONTRACE( " turning around... " );
	      skills->exec(worldSnap, tcmd);
	      thcmd = skills->turnHead(worldSnap, tcmd.bodyDir);
	      wantToTurnHead = true;
	    }
	    else {  // should never happern
	      DECISIONTRACE( " I allready turned around??? " );
	      skills->noAction();
	      lookAround(skills, worldSnap, worldSnap->ball.pos);
	    }
	  }
	  else {
	    skills->noAction();
	    lookAround(skills, worldSnap, myPos);
	  }
	}
      }
      else { // i will not kick
	skills->noAction();
	lookAround(skills, worldSnap, ball);
      }
    }

  OTHERTRACE("waitingTime: " << waitingTime);
  if (waitingTime < 0 &&
      worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 != 0)
    waitingTime = 0;

  return true;
}

bool 
StandardSituationModule::
doingKickInUs(AdvSkills *skills,
	      const Snapshot *worldSnap,
	      const int players[],
	      const Vector2d target[],
	      const int count)
{
  Vector2d targetPos;
  bool iWillKick(false);
  double r(BALL_SIZE), mySize(worldSnap->my.size);
  Vector2d myPos(worldSnap->my.pos), ball(worldSnap->ball.pos);
  
  for (int i = 0; i < count; i++)
    if (worldSnap->my.number == players[i] )
      {
	DECISIONTRACE( " I am player " << i+1 << " out of " << count << " players");
	targetPos = target[i];
	if (i == 0) iWillKick = true;
      }

  DECISIONTRACE( " My targetPos: " << targetPos );
  OTHERTRACE( " My pos: " << myPos );

  if (iWillKick) { //[C]
    r = mySize + worldSnap->my.kickableMargin;
    if (targetPos.getY() > ball.getY())
      targetPos += Vector2d(0, r+PLAYER_SIZE);
    else
      targetPos -= Vector2d(0, r+PLAYER_SIZE);

    if (worldSnap->my.inCorridor(ball, targetPos, 4*mySize, PI_2)) {
      DRAWTRACE( CORRIDOR(ball, targetPos, 4*mySize, PI_2) );
      targetPos = target[0];
      r = BALL_SIZE;
    }
    DRAWTRACE( POSITION(targetPos) );
  }

  double stamina(calcUsableStamina(activeStaminaFactor, worldSnap));
  
  if (!goingTo(targetPos, stamina, skills, 
	       worldSnap, &ball, r))
    {
      if (iWillKick) {
	bool wait(true);
	int myNum(worldSnap->my.number);
	Grade passGrade;

	for (int pnum = 1; pnum < 12; pnum++)
	  if (pnum != myNum) {
	    passGrade = passEval->ourEval(worldSnap, myNum, pnum);
	    if (passGrade.is_better_than(Grade::AVERAGE) &&
		worldSnap->ourTeam(pnum).age <= 2) 
	      wait = false;

	    DRAWTRACE( LINE(ball, worldSnap->ourTeam(pnum).pos) );
#ifdef USEDRAWTRACE
	    {
	      ostrstream text;
	      text << passGrade << (char)0;
	      DRAWTRACE( TEXT(text.str(), worldSnap->ourTeam(pnum).pos) );
	    }
#endif
	  }
	  else
	    continue;

	OTHERTRACE("next time of EMRG channel: " << worldSnap->nextTimeOfChannel(AUMUL_EMRG));

	if (((waitingTime-- <= 0 || !wait)
	     && worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 == 0)
	    || timeLeft <= 0) {
	  stdSitFinished = true;
	  AdvSkills::TurnCommand tcmd = 
	    skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
			 worldSnap->my.bodyDir);
	  if (tcmd.state != AdvSkills::RFINISHED)
	    skills->exec(worldSnap, tcmd);
	  else
	    skills->noAction();
	  thcmd = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
	  wantToTurnHead = true;
	  DECISIONTRACE("Standardsituation finished");
	}
	else {
	  skills->noAction();
	  lookAround(skills, worldSnap, myPos);
	}
      }
      else { // i will not kick
	skills->noAction();
	lookAround(skills, worldSnap, ball);
      }
    }    
    
  OTHERTRACE("waitingTime: " << waitingTime);
  if (waitingTime < 0 &&
      worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 != 0)
    waitingTime = 0;

  return true;
}

bool 
StandardSituationModule::
doingCornerKickUs(AdvSkills *skills,
		  const Snapshot *worldSnap,
		  const int players[],
		  const Vector2d target[],
		  const int count)
{
  Vector2d targetPos;
  bool iWillKick(false);
  double r(BALL_SIZE), mySize(worldSnap->my.size);
  Vector2d myPos(worldSnap->my.pos), ball(worldSnap->ball.pos);
  
  for (int i = 0; i < count; i++)
    if (worldSnap->my.number == players[i] )
      {
	DECISIONTRACE( " I am player " << i+1 << " out of " << count << " players");
	targetPos = target[i];
	if (i == 0) iWillKick = true;
      }

  DECISIONTRACE( " My targetPos: " << targetPos );
  OTHERTRACE( " My pos: " << myPos );

  if (iWillKick) { //[C]
    r = mySize + worldSnap->my.kickableMargin;
    if (targetPos.getY() > 0) {
      if (targetPos.getX() < 0)
	targetPos += Vector2d(-(r+PLAYER_SIZE), r+PLAYER_SIZE);
      else 
	targetPos += Vector2d(r+PLAYER_SIZE, r+PLAYER_SIZE);
    }
    else {
      if (targetPos.getX() < 0)
	targetPos -= Vector2d(r+PLAYER_SIZE, r+PLAYER_SIZE);
      else 
	targetPos -= Vector2d(-(r+PLAYER_SIZE), r+PLAYER_SIZE);
    }
    if (worldSnap->my.inCorridor(ball, targetPos, 4*mySize, PI_2)) {
      DRAWTRACE( CORRIDOR(ball, targetPos, 4*mySize, PI_2) );
      targetPos = target[0];
      r = BALL_SIZE;
    }
    DRAWTRACE( POSITION(targetPos) );
  }

  double stamina(calcUsableStamina(activeStaminaFactor, worldSnap));
  
  if (!goingTo(targetPos, stamina, skills, 
	       worldSnap, &ball, r))
    {
      if (iWillKick) {
	bool wait(true);
	int myNum(worldSnap->my.number);
	Grade passGrade;

	for (int pnum = 1; pnum < 12; pnum++)
	  if (pnum != myNum) {
	    passGrade = passEval->ourEval(worldSnap, myNum, pnum);
	    if (passGrade.is_better_than(Grade::AVERAGE) &&
		worldSnap->ourTeam(pnum).age <= 2) 
	      wait = false;

	    DRAWTRACE( LINE(ball, worldSnap->ourTeam(pnum).pos) );
#ifdef USEDRAWTRACE
	    {
	      ostrstream text;
	      text << passGrade << (char)0;
	      DRAWTRACE( TEXT(text.str(), worldSnap->ourTeam(pnum).pos) );
	    }
#endif
	  }
	  else
	    continue;

	OTHERTRACE("next time of EMRG channel: " << worldSnap->nextTimeOfChannel(AUMUL_EMRG));

	if (((waitingTime-- <= 0 || !wait)
	     && worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 == 0)
	    || timeLeft <= 0) {
	  stdSitFinished = true;
	  AdvSkills::TurnCommand tcmd = 
	    skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
			 worldSnap->my.bodyDir);
	  if (tcmd.state != AdvSkills::RFINISHED)
	    skills->exec(worldSnap, tcmd);
	  else
	    skills->noAction();
	  thcmd = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
	  wantToTurnHead = true;
	  DECISIONTRACE("Standardsituation finished");
	}
	else {
	  skills->noAction();
	  lookAround(skills, worldSnap, myPos);
	}
      }
      else { // i will not kick
	skills->noAction();
	lookAround(skills, worldSnap, ball);
      }
    }    

  OTHERTRACE("waitingTime: " << waitingTime);
  if (waitingTime < 0 &&
      worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 != 0)
    waitingTime = 0;
    
  return true;
}

bool 
StandardSituationModule::
doingKickOff(AdvSkills *skills,
	     const Snapshot *worldSnap,
	     TacticalSnapshot *tacticalSnap)

{
  // moving to standard kickoff-positions and turning after that
  Vector2d moveTarget;
  moveTarget = calcKickOffTarget(worldSnap->my.number, worldSnap, tacticalSnap);
  // [C]*
  if ((worldSnap->my.distTo(moveTarget) > 0.4) || 
      (worldSnap->my.pos.getX() > -0.1) )
    {
      if ((worldSnap->playMode == PLAY_MODE_BEFORE_KICK_OFF) ||
	  (worldSnap->playMode == PLAY_MODE_GOAL_US) ||
	  (worldSnap->playMode == PLAY_MODE_GOAL_THEM) 
	  )
	{
	  skills->basicMove(moveTarget.getX(),moveTarget.getY());
	  DECISIONTRACE( "doingKickOff: moves the player to the kickoff position at " 
			 << moveTarget );
	  return true;
	}      
      // [C]
      if (worldSnap->my.distTo(moveTarget) > 0.8)
	{
	  DECISIONTRACE( "doingKickOff: moves not possible, walking to" 
			 << moveTarget );
	  // [C]
	  skills->exec(worldSnap, skills->goToPosition(worldSnap, moveTarget, 0.8));
	  return true;
	}
    }
  if (worldSnap->playMode == PLAY_MODE_BEFORE_KICK_OFF)
    {
      skills->basicTurn(M_PI/4.0);
      DECISIONTRACE( "doingKickOff: turns the player (cause we are the MRB)" );
      return true;
    }

  if (!weAreKickingOff)
    {
      DECISIONTRACE( "doingKickOff: otherTeam is kicking off, so turning to ball" );
      skills->basicTurn(worldSnap->my.bodyAngleTo(worldSnap->ball));
      return true;
    }  
   
  if (!skills->kickSafe(worldSnap))
    {
      DECISIONTRACE( "I am not responsible for the kick off, so turning to ball" );
      skills->basicTurn(worldSnap->my.bodyAngleTo(worldSnap->ball));
      return true;
    }
   
  // I am responsible for the kick-off.
  // Check if it is neccesary to wait. (other players should turn to the ball 
  // before I do the kick off)
  if ((worldSnap->time - standardWaitingTime > lastTime_KICK_OFF_US - 5)
      && (waitingTime) < 0)    
    waitingTime = 3;

  // find out where to kick to 
  int passReceiver;
  Vector2d kickTarget = calcKickOffKickTarget(worldSnap, tacticalSnap, passReceiver);

  if (waitingTime > 0)
    {
      DECISIONTRACE("I am responsible for the ball, Waiting for the Kickoff, waitingTime = "
		    << waitingTime);
      // DECISIONTRACE("=> turning to the others goal");
      // skills->basicTurn(worldSnap->my.bodyAngleTo(worldSnap->otherGoal.pos));
      DECISIONTRACE("=> turning to the player I want to pass to");
      skills->basicTurn(worldSnap->my.bodyAngleTo(kickTarget));

      waitingTime --;
      return true;
    }

  DECISIONTRACE( "I am responsible for the kick off, kicking to " 
		 << kickTarget );     
  // [C]
  AdvSkills::ExtKickCommand kickCommand = skills->kick(worldSnap, 0.6, 
						       kickTarget, 1);
  
  skills->exec(worldSnap, kickCommand);

  // report tactical action 
  tacticalSnap->reportOurTacticalAction(TA_PASS, worldSnap->my.number, passReceiver);
  return true;
}

Vector2d 
StandardSituationModule::
calcKickOffKickTarget(const Snapshot *worldSnap, 
		      const TacticalSnapshot *tacticalSnap,
		      int &passReceiver) const
{
  int ourSystem = getOurTacticalSystem(tacticalSnap);  
  // [C]
  int tacticalPassReceiver = 7; // the player I want to pass to (tactical number !!)

  passReceiver = tacticalPassReceiver;           // to avoid warnings

  // check what positions the other team has
  int upperY=0;
  int lowerY=0; 
  for (int i=1; i<=11; i++) 
    { 
      if (worldSnap->otherTeam(i).age > maxAge)
	continue;
      // [C]
      if (worldSnap->otherTeam(i).pos.getX() > FIELD_LENGTH / 2.0)  
	continue; // players far away from the middle line don't matter
      // [C]
      if (worldSnap->otherTeam(i).pos.getY() < FIELD_WIDTH / 20.0)
	lowerY++; // players in the lower half
      // [C]
      if (worldSnap->otherTeam(i).pos.getY() > -FIELD_WIDTH / 20.0)
	upperY++; // players in the upper half
    }

  switch(ourSystem)
    {
      // [C]*
    case 442:
      if (upperY > lowerY)
	tacticalPassReceiver = 8;
      else if (upperY < lowerY)
	tacticalPassReceiver = 7;
      else
	tacticalPassReceiver = 10;
      break;
    case 343:
      tacticalPassReceiver = 7;
      break;
    case 352:
      tacticalPassReceiver = 7;
      break;
    default:
      passReceiver = 7;
      ERRORTRACE( "unknown tactical system, kicking into the crowd !!" );
      break;
    }
  DECISIONTRACE( "tactical system: " << ourSystem << " kick for player " 
		 << tacticalPassReceiver );


  // now get the player's position
  passReceiver = tacticalPassReceiver;
  for (int i=1; i<=11; i++)
    if (tacticalSnap->ourTeamTacticalNo(i) == tacticalPassReceiver)
      {
	passReceiver = i;
	break;
      }

  if (worldSnap->ourTeam(passReceiver).age < maxAge)
    return worldSnap->ourTeam(passReceiver).pos;
  else
    // [C]
    return Vector2d(-10.0, 0.0);
}

Grade
StandardSituationModule::
evalRemainingKicksUs(AdvSkills* skills,
		     const Snapshot* worldSnap, 
		     const TacticalSnapshot* tacticalSnap, 
		     const play_mode playMode)
{
  Grade value;

  timeLeft--;
  switch(playMode)
    {
    case PLAY_MODE_KICK_IN_US : 
      actCount = calcPlayersAndPos4KickInUs(worldSnap, tacticalSnap, 
					    actPlayer, targetPos);
      break;
    case PLAY_MODE_CORNER_KICK_US :
      actCount = calcPlayersAndPos4CornerKickUs(worldSnap, tacticalSnap, 
						actPlayer, targetPos);
      break;
    default: ;
    }

  for (int i = 0; i < actCount; i++) {
    if (worldSnap->my.number == actPlayer[i])
      {
	value.set_ABSOLUTE();
	// max time to wait for second player to arrive its position
	if (waitingTime < 0)
	  switch(playMode)
	    {
	    case PLAY_MODE_KICK_IN_US:
	      waitingTime = standardWaitingTime;
	      break;
	    case PLAY_MODE_CORNER_KICK_US:
	      waitingTime = standardWaitingTime / 2; // [C]
	      break;
	    default: ;
	    }
	break;
      }
    else
      {
	if (mustAvoidBall(worldSnap))
	  value.set_ABSOLUTE();
	else
	  value.set_BAD();
      }
  }

  ensure(value.is_defined());

  return value;
}

Grade
StandardSituationModule::
evalFreeKickUs(AdvSkills* skills,
	       const Snapshot* worldSnap, 
	       const TacticalSnapshot* tacticalSnap)
{
  if (skills);      // to avoid warnings

  Grade value;
    
  // if I am the goalie and want another module to act, do nothing and
  // return BAD

  if (iAmGoalie && nextCycleDoNothing)
    {
      nextCycleDoNothing = false;
      value.set_BAD();
      DECISIONTRACE( "Goalie wants the other modules to kick away the ball" );
      return value;
    }
  
  // If I am the goalie and have catched the ball => move and kick
  bool goalieHasCatched  =  
    (worldSnap->ourGoalieNo != UNKNOWN) && 
    // [C]
    ( sqrnorm(worldSnap->ourTeam(worldSnap->ourGoalieNo).pos - worldSnap->ball.pos) < 1.0 );
  
  if (iAmGoalie)
    goalieHasCatched = goalieHasCatched && (lastTime_PLAY_ON <= worldSnap->lastCatchTime);

  if ( goalieHasCatched )
    {
      DECISIONTRACE( "goalie has catched the ball in cycle" 
		     << worldSnap->lastCatchTime
		     << " lastTime_PLAY_ON= " << lastTime_PLAY_ON
		     << " so doing goalieMove&Kick" );
      goalieDoesFreeKick = true;
      goalieMayMove = true;
    }
  else
    {
      DECISIONTRACE( "goalie has not catched the ball (last catch in cycle " 
  		     <<  worldSnap->lastCatchTime << ") so normal freekick" );
//        DECISIONTRACE( "goalie has not catched the ball (last catch in cycle " 
//  		     <<  worldSnap->lastCatchTime << ") so normal goalieMove&Kick" );
//        // [C]*
//        if ( (worldSnap->ball.pos.getX() < -FIELD_LENGTH/2. + PENALTY_AREA_LENGTH -1.0 ) &&
//  	   (fabs(worldSnap->ball.pos.getY()) < PENALTY_AREA_WIDTH / 2.0 - 1.0) )
//  	{
//  	  DECISIONTRACE( "freekick in our penalty area, goalie should do it" );
//  	  goalieDoesFreeKick = true;
//  	  goalieMayMove = false;
//  	  goalieShouldMoveAgain = false;
//  	}
//        else
	goalieDoesFreeKick = false;
    }
  
  if (goalieDoesFreeKick)
    {
      if (iAmGoalie)
	{
	  value.set_ABSOLUTE();
	  if (waitingTime < 0)
	    waitingTime = standardWaitingTime;
	}
      else 
	{
	  // give players near the ball no chance to kick the ball
	  // better let them walk away
	  if (mustAvoidBall(worldSnap) || worldSnap->my.inOurPenaltyArea())
	    value.set_ABSOLUTE();
	  else
	    value.set_IMPOSSIBLE();
	}
      return value;
    }

   
  // goalie does not do the freekick
  if ( iAmGoalie &&
       worldSnap->ball.inOurPenaltyArea(1) )
    {
      // our freekick in our penalty area, so better get the goalie back towards our 
      // goal
      DECISIONTRACE( "Goalie wants to run back to our goal" );
      value.set_ABSOLUTE();
      return value;
    }


  // if we are close to the edge of the field
  // [C]
  if (fabs(fabs(worldSnap->ball.pos.getY())-FIELD_WIDTH/2.0) < 2.0)
    //do a kick in
    actCount = calcPlayersAndPos4KickInUs(worldSnap, tacticalSnap, 
					  actPlayer, targetPos);
  else
    // do a free kick
    actCount = calcPlayersAndPos4FreeKickUs(worldSnap, tacticalSnap,
					    actPlayer, targetPos);

  for (int i = 0; i < actCount; i++) {
    if (worldSnap->my.number == actPlayer[i])
      {     
	value.set_ABSOLUTE();

	// max time to wait for second player to arrive its position
	if (waitingTime < 0)
	  waitingTime = standardWaitingTime;
	return value;
      }
    }

  DECISIONTRACE( "I am not responsible for ball" );
  if (mustAvoidBall(worldSnap))
    value.set_ABSOLUTE();
  else
    value.set_BAD();

  ensure(value.is_defined());

  return value;
}

Vector2d 
StandardSituationModule::
calcKickOffTarget(int playerNumber,
		  const Snapshot *worldSnap,
		  const TacticalSnapshot *tacticalSnap) const
{
  if (worldSnap);     // to avoid warnings

  int tacticalNumber 
    = tacticalSnap->ourTeamTacticalNo(playerNumber);
  if (tacticalNumber == UNKNOWN)
    tacticalNumber = playerNumber;

  int ourSystem = getOurTacticalSystem(tacticalSnap);

  switch(ourSystem)
    {
    case 442:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_442[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_442[tacticalNumber-1];
    case 343:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_343[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_343[tacticalNumber-1];
    case 352:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_352[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_352[tacticalNumber-1];
    default: // 343
      ERRORTRACE( "unknown tactical system: " << ourSystem );
      if (weAreKickingOff)
	return kickOffPosWeKickOff_343[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_343[tacticalNumber-1];
      
    }
}

bool 
StandardSituationModule::
doingFieldPlayerGoalKickUs(AdvSkills *skills,
			   const Snapshot *worldSnap,
			   TacticalSnapshot *tacticalSnap)
{
  // check if we have to leave the penaltyarea and do it
  bool getOutOfPenaltyArea(false);
  Vector2d dest = targetPos[worldSnap->my.number-1]; 

  switch(getOurTacticalSystem(tacticalSnap))
    {
    case 442:
      if  (!( (worldSnap->my.number == 2) ||
	      (worldSnap->my.number == 5) ) )
	getOutOfPenaltyArea = true;
      break;
    case 343:
    case 352:
    default:
      if  (!( (worldSnap->my.number == 2) ||
	      (worldSnap->my.number == 4) ) )
	getOutOfPenaltyArea = true;
      break;
    }

  if (getOutOfPenaltyArea)
    {
      Vector2d posEvalTarget =
	positionEval->GetPlayerPosition(worldSnap->my.number);
      WorldObject wO;
      wO.pos = posEvalTarget;
      if (wO.inOurPenaltyArea())
	wO.pos.setX(-FIELD_LENGTH/2. + PENALTY_AREA_LENGTH + 8.);
      AdvSkills::TurtleCommand tcmd =
	skills->goToPosition(worldSnap, wO.pos, 1);
      if (tcmd.state != AdvSkills::TFINISHED)
	skills->exec(worldSnap, tcmd);
      else
	skills->basicTurn(PI/4.);
      lookAround(skills, worldSnap, worldSnap->ball.pos);
      return true;
    }

  // if not yet reached go to wanted positon and wait for ball
  Vector2d leftPos(-FIELD_LENGTH/2.+GOAL_AREA_LENGTH, -GOAL_AREA_WIDTH/2.), 
    rightPos(-FIELD_LENGTH/2.+GOAL_AREA_LENGTH, GOAL_AREA_WIDTH/2.);
  bool goalieShot = 
    sqrnorm(leftPos-worldSnap->ball.pos) > 4 &&
    sqrnorm(rightPos-worldSnap->ball.pos) > 4;
  bool going(false);


  if (!goalieShot &&
      (going = goingTo(dest, STAMINA_INC_MAX, skills, worldSnap)) )
    {
      DECISIONTRACE( "going to " << dest );
      return true;
    }

  // check if we need to intercept ball
  bool icpt(false);

  if (goalieShot && !skills->kickSafe(worldSnap)) {
    WorldBall ball = worldSnap->ball;
    Vector2d myPos = worldSnap->my.pos;

    int numberOfDashes(3);
    // ??? PLAYER_DECAY HAS TO be changed for het. players
    double reachDist = MAXPOWER*worldSnap->my.dashPowerRate/(1 - PLAYER_DECAY)*
      (numberOfDashes - PLAYER_DECAY*(1 - power(PLAYER_DECAY, numberOfDashes))
       /(1 - PLAYER_DECAY));
      
    double yPos;
    if (ball.pos.getY() > 0) 
      yPos = GOAL_AREA_WIDTH/2;
    else
      yPos = -GOAL_AREA_WIDTH/2.;

    Vector2d goalAreaPos(-FIELD_LENGTH/2. + GOAL_AREA_LENGTH, yPos);
    double ang = atan(reachDist/worldSnap->my.distTo(goalAreaPos));

    if (!ball.inCorridor(goalAreaPos, myPos, reachDist, ang))
      icpt = true;
    else {
      // if ball near to me, intercept it
      if ( (worldSnap->my.distTo(worldSnap->ball.pos) < reachDist*1.2) ) //[C]
 	icpt = true;
    }
  }

  // check if I want to turn to the ball
  if (goalieShot && !icpt && !skills->kickSafe(worldSnap)) {
    AdvSkills::TurnCommand tcmd =
      skills->turn(worldSnap, 
		   worldSnap->my.bodyAngleTo(worldSnap->ball)+
		   worldSnap->my.bodyDir);
    
    if (tcmd.state != AdvSkills::RFINISHED) {
      skills->exec(worldSnap, tcmd);
      thcmd = skills->turnHead(worldSnap, worldSnap->ball);
      DECISIONTRACE("turning to ball");
      thcmd.vwidth = VIEW_WIDTH_NARROW;
      wantToTurnHead = true;
      return true;
    }
  }
  
  // we are not going and don`t need to intercept, so wait
  if (!going && !icpt && !skills->kickSafe(worldSnap)) {
    DECISIONTRACE( "destination "  << dest << " reached" );
    DECISIONTRACE( "and waiting" );
    // if in upper field turn vector left else right
    skills->noAction();
    int vz( (worldSnap->my.pos.getY() > 0) ? 1 : -1 );
    lookAround(skills, worldSnap, 
	       (worldSnap->ball.pos-worldSnap->my.pos).
	       turn(vz*DEG_TO_RAD(lookAroundAngle-vz*5.))+
	       worldSnap->my.pos);
    return true;
  }

  // we need to icpt
  if (icpt)
    {
      WorldBall dummyBall;
      AdvSkills::TurtleCommand tcmd =
	skills->controlBall(worldSnap, worldSnap->senseBody.stamina, 
			    0.2*KICKABLE_MARGIN, dummyBall);
      if (tcmd.state != AdvSkills::TFINISHED)
	{
	  skills->exec(worldSnap, tcmd);
	  DECISIONTRACE( "controlling ball, meeting @ "
			 << dummyBall.pos << "dist: "
			 << norm(dummyBall.pos-worldSnap->my.pos));
	  thcmd = skills->turnHead(worldSnap, worldSnap->ball);
	  thcmd.vwidth = VIEW_WIDTH_NARROW;
	  wantToTurnHead = true;
	  return true;
	}
    }

  // we intercepted, so go on ...
  if (skills->kickSafe(worldSnap))
    {
      AdvSkills::TurnCommand tcmd;
      AdvSkills::KickCommand kcmd;

      DECISIONTRACE( "ball in kickrange" );

      Vector2d placePos;
      placePos.setRadAng(skills->safeCircle, 0.); //[C]
      DRAWTRACE( POSITION(placePos) );
	   
      // first stop the ball
      kcmd = skills->stopBall(worldSnap);
      if (kcmd.state == AdvSkills::KSTOPPING && 
	  sqrnorm(worldSnap->ball.v) > .05) {
	skills->exec(worldSnap, kcmd);
	DECISIONTRACE( "stopping ball" );
	thcmd = skills->turnHead(worldSnap, worldSnap->ball);
	thcmd.vwidth = VIEW_WIDTH_NARROW;
	wantToTurnHead = true;
	return true;
      }

      // now turn towards other goal
      tcmd = skills->turn(worldSnap, 
			  worldSnap->my.bodyAngleTo(placePos)+
			  worldSnap->my.bodyDir);

      if (tcmd.state != AdvSkills::RFINISHED) {
	skills->exec(worldSnap, tcmd);
	DECISIONTRACE("turning to placePos");
	thcmd = skills->turnHead(worldSnap, worldSnap->ball);
	thcmd.vwidth = VIEW_WIDTH_NARROW;
	wantToTurnHead = true;
	return true;
      }
      
      // now place ball in front of us
      kcmd = skills->placeBall(worldSnap, placePos);
      if (kcmd.state == AdvSkills::KPLACING
	  && sqrnorm(kcmd.v) > .05) { //[C]
	skills->exec(worldSnap, kcmd);
	DECISIONTRACE("placing ball");
	thcmd = skills->turnHead(worldSnap, worldSnap->ball);
	thcmd.vwidth = VIEW_WIDTH_NARROW;
	wantToTurnHead = true;
	return true;
      }

      // ball in front of us, so collect information
      skills->noAction();
      lookAround(skills, worldSnap, placePos);
      waitingTime--;
      OTHERTRACE("waitingTime: " << waitingTime);
      DECISIONTRACE("turned to ball. looking around");

      if ((waitingTime <= 0 &&
	   worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-1 == 0) ||
	  timeLeft <= 0) {
	  stdSitFinished = true;
	  DECISIONTRACE("Standardsituation finished");
	  return true;
      }

      if (waitingTime <= 0)
	waitingTime = 1;

      return true;
    }
  else
    {
      ERRORTRACE("Ball not kickable after trying to controll it!!!");
      skills->noAction();
    }
	    
  return false;
}
