//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// \$Id:
// \$manager: gauer$
//
//////////////////////////////////////////////////////////////////////


#include "standardsituationmodule.h"
#include "Config.h"


bool AfterStandardSituationPlan::executePlan(AdvSkills *skills, const Snapshot *worldSnap)
{
  switch (thePlan)
    {
    case PLAN_SUPERKICK:
      {
	if (time > 0)
	  {
	    time --;
	    if (skills->kickSafe(worldSnap))
	      {
		kickCommand = skills->rekick(worldSnap, kickCommand);
		skills->exec(worldSnap, kickCommand);
		PLANTRACE( " Execute Superkick, time = " << time );
		return true; 
	      }	    
	    PLANTRACE( " planned Superkick seems to be ready (ball not kickable), time =  "
		       << time );
	  }
	else
	  PLANTRACE ( " planned Superkick interrupted because of time-out " );
	setNothing();
	return false; 
      }     
      
    case PLAN_RUN_TO_POSITION:
      {
	if (time > 0)
	  {
	    time --;
	    double usedStam = max(stamina, STAMINA_INC_MAX);
	    AdvSkills::TurtleCommand turtle;
	    turtle = skills->goToPosition(worldSnap, pos, 2.0, usedStam);
	    if (turtle.state == AdvSkills::TFINISHED)
	      {
		PLANTRACE( " Do not execute planned RunToPosition, because destination is reached ");
		setNothing();
		return false;
	      }
	    PLANTRACE( " Execute planned RunToPosition to " << pos << " with " 
		       << usedStam << " stamina, " 
		       << " my.pos = " << worldSnap->my.pos
		       << ", time = " << time);
	    skills->exec(worldSnap, turtle);
	    return true;
	  }
	PLANTRACE( "RunToPosition stopped because of time-out" );
	setNothing();
	return false;
    }      
    
    case PLAN_NO_ACTION:
      if (time > 0)
	{
	  skills->noAction();
	  time --;
	  PLANTRACE( " Executing planned noAction, time =  " << time );
	  if (time == 0)
	    setNothing();
	  return true;
	}
      PLANTRACE( " End of doing nothing " );
      return false;
 
    case PLAN_TURNING:
      if (time > 0)
	{
	  skills->exec(worldSnap, skills->turn(worldSnap, dir));
	  time --;
	  PLANTRACE( " Executing planned turn to " << dir << ", time = " << time );
	  if (time == 0)
	    setNothing();
	  return true;
	}
      PLANTRACE( " End of turning " );
      return false;

    case PLAN_NOTHING:
    default:
      PLANTRACE( " UNKNOWN PLAN !?! GRRRRRRRRR " );
      ERRORTRACE( cerr << " UNKNOWN PLAN !?! GRRRRRRRRR " << endl );
      return false;
    }
}


